import datetime
import pathlib
from typing import List, Set

# Needed to import `tools` for local testing
from femr_test_tools import EventsWithLabels, run_test_for_labeler

from femr.labelers import TimeHorizon
from femr.labelers.omop import (
    AKICodeLabeler,
    AnemiaCodeLabeler,
    CodeLabeler,
    HyperkalemiaCodeLabeler,
    HypoglycemiaCodeLabeler,
    HyponatremiaCodeLabeler,
    LupusCodeLabeler,
    MortalityCodeLabeler,
    NeutropeniaCodeLabeler,
    OMOPConceptCodeLabeler,
    ThrombocytopeniaCodeLabeler,
)

#############################################
#############################################
#
# Generic CodeLabeler
#
#############################################
#############################################


def test_outcome_codes(tmp_path: pathlib.Path):
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=10))
    # One outcome
    labeler = CodeLabeler(["2"], time_horizon)
    events_with_labels: EventsWithLabels = [
        (((2015, 1, 3), 2, None), "duplicate"),
        (((2015, 1, 3), 4, None), "duplicate"),
        (((2015, 1, 3), 1, None), "duplicate"),
        (((2015, 1, 3), 3, None), "skip"),
        (((2015, 10, 5), 1, None), False),
        (((2018, 1, 3), 2, None), "skip"),
        (((2018, 3, 1), 4, None), False),
        (((2018, 3, 3), 1, None), False),
        (((2018, 5, 2), 5, None), True),
        (((2018, 5, 3), 2, None), "skip"),
        (((2018, 5, 3, 11), 1, None), False),
        (((2018, 5, 4), 1, None), "duplicate"),
        (((2018, 5, 4), 4, None), False),
        (((2018, 11, 1), 5, None), False),
        (((2018, 12, 4), 1, None), False),
        (((2018, 12, 30), 4, None), "out of range"),
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_outcome_codes_one")

    # Zero outcome
    labeler = CodeLabeler([], time_horizon)
    events_with_labels = [
        (((2015, 1, 3), 2, None), "duplicate"),
        (((2015, 1, 3), 4, None), "duplicate"),
        (((2015, 1, 3), 1, None), "duplicate"),
        (((2015, 1, 3), 3, None), False),
        (((2015, 10, 5), 1, None), False),
        (((2018, 1, 3), 2, None), False),
        (((2018, 3, 1), 4, None), False),
        (((2018, 3, 3), 1, None), False),
        (((2018, 5, 2), 5, None), False),
        (((2018, 5, 3), 2, None), False),
        (((2018, 5, 3, 11), 1, None), False),
        (((2018, 5, 4), 1, None), "duplicate"),
        (((2018, 5, 4), 4, None), False),
        (((2018, 11, 1), 5, None), False),
        (((2018, 12, 4), 1, None), False),
        (((2018, 12, 30), 4, None), "out of range"),
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_outcome_codes_zero")

    # Multiple outcomes
    labeler = CodeLabeler(["1", "4"], time_horizon)
    events_with_labels = [
        (((2015, 1, 3), 2, None), "duplicate"),
        (((2015, 1, 3), 4, None), "duplicate"),
        (((2015, 1, 3), 1, None), "duplicate"),
        (((2015, 1, 3), 3, None), "skip"),
        (((2015, 10, 5), 1, None), "skip"),
        (((2018, 1, 3), 2, None), False),
        (((2018, 3, 1), 4, None), "skip"),
        (((2018, 3, 3), 1, None), "skip"),
        (((2018, 5, 2), 5, None), False),
        (((2018, 5, 3), 2, None), False),
        (((2018, 6, 2), 0, None), True),
        (((2018, 6, 3, 11), 1, None), "skip"),
        (((2018, 6, 3, 23), 3, None), False),
        (((2018, 9, 1), 3, None), True),
        (((2018, 9, 4), 4, None), "skip"),
        (((2018, 11, 1), 5, None), False),
        (((2018, 12, 3), 0, None), True),
        (((2018, 12, 4), 4, None), "skip"),
        (((2018, 12, 30), 0, None), "out of range"),
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="test_outcome_codes_multiple")


def test_prediction_codes(tmp_path: pathlib.Path):
    # One outcome + multiple predictions
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=10))
    labeler = CodeLabeler(["2"], time_horizon, prediction_codes=["4", "5"])
    events_with_labels: EventsWithLabels = [
        (((2015, 1, 3), 2, None), "skip"),
        (((2015, 1, 3), 4, None), "skip"),
        (((2015, 1, 3), 1, None), "skip"),
        (((2015, 1, 3), 3, None), "skip"),
        (((2015, 10, 5), 1, None), "skip"),
        (((2018, 1, 3), 2, None), "skip"),
        (((2018, 3, 1), 4, None), False),
        (((2018, 3, 3), 1, None), "skip"),
        (((2018, 5, 2), 5, None), True),
        (((2018, 5, 3), 2, None), "skip"),
        (((2018, 5, 3, 11), 1, None), "skip"),
        (((2018, 5, 4), 4, None), False),
        (((2018, 5, 4), 1, None), "skip"),
        (((2018, 11, 1), 5, None), False),
        (((2018, 12, 4), 1, None), "skip"),
        (((2018, 12, 30), 4, None), "out of range"),
    ]
    run_test_for_labeler(labeler, events_with_labels, help_text="prediction_codes_one_outcomes")

    # Multiple outcomes + multiple predictions
    labeler = CodeLabeler(["2", "6", "7"], time_horizon, prediction_codes=["4", "5"])
    events_with_labels = [
        (((2010, 1, 1), 2, None), "skip"),
        (((2010, 1, 3), 4, None), True),
        (((2010, 1, 8), 6, None), "skip"),
        (((2010, 2, 1), 5, None), True),
        (((2010, 2, 9), 7, None), "skip"),
        (((2010, 2, 11), 4, None), False),
        (((2015, 1, 3), 2, None), "skip"),
        (((2015, 1, 3), 4, None), "skip"),
        (((2015, 1, 3), 1, None), "skip"),
        (((2015, 1, 3), 3, None), "skip"),
        (((2015, 10, 5), 1, None), "skip"),
        (((2018, 1, 3), 2, None), "skip"),
        (((2018, 3, 1), 4, None), True),
        (((2018, 3, 2), 7, None), "skip"),
        (((2018, 3, 3), 1, None), "skip"),
        (((2018, 5, 2), 5, None), True),
        (((2018, 5, 3), 2, None), "skip"),
        (((2018, 5, 3, 11), 1, None), "skip"),
        (((2018, 5, 4), 4, None), False),
        (((2018, 5, 4), 1, None), "skip"),
        (((2018, 11, 1), 5, None), False),
        (((2018, 12, 4), 1, None), "skip"),
        (((2018, 12, 30), 4, None), "out of range"),
    ]
    run_test_for_labeler(
        labeler,
        events_with_labels,
        help_text="prediction_codes_multiple_outcomes",
    )

    # Multiple outcomes + no predictions
    labeler = CodeLabeler(["2", "6", "7"], time_horizon, prediction_codes=[])
    events_with_labels = [
        (((2010, 1, 1), 2, None), "skip"),
        (((2010, 1, 3), 4, None), "skip"),
        (((2010, 1, 8), 6, None), "skip"),
        (((2010, 2, 1), 5, None), "skip"),
        (((2010, 2, 9), 7, None), "skip"),
        (((2010, 2, 11), 4, None), "skip"),
        (((2015, 1, 3), 2, None), "skip"),
        (((2015, 1, 3), 4, None), "skip"),
        (((2015, 1, 3), 1, None), "skip"),
        (((2015, 1, 3), 3, None), "skip"),
        (((2015, 10, 5), 1, None), "skip"),
        (((2018, 1, 3), 2, None), "skip"),
        (((2018, 3, 1), 4, None), "skip"),
        (((2018, 3, 2), 7, None), "skip"),
        (((2018, 3, 3), 1, None), "skip"),
        (((2018, 5, 2), 5, None), "skip"),
        (((2018, 5, 3), 2, None), "skip"),
        (((2018, 5, 3, 11), 1, None), "skip"),
        (((2018, 5, 4), 4, None), "skip"),
        (((2018, 5, 4), 1, None), "skip"),
        (((2018, 11, 1), 5, None), "skip"),
        (((2018, 12, 4), 1, None), "skip"),
        (((2018, 12, 30), 4, None), "skip"),
    ]
    run_test_for_labeler(
        labeler,
        events_with_labels,
        help_text="prediction_codes_zero_predictions",
    )


#############################################
#############################################
#
# Generic OMOPConceptCodeLabeler
#
#############################################
#############################################


class DummyOntology_Base:
    def get_children(self, code: str) -> Set[str]:
        return set()

    def get_all_children(self, code: str) -> Set[str]:
        val = {code}
        for child in self.get_children(code):
            val |= self.get_all_children(child)
        return val


class DummyOntology_OMOPConcept(DummyOntology_Base):
    def get_children(self, parent_code: str) -> Set[str]:
        if parent_code == "OMOP_CONCEPT_A":
            return {"OMOP_CONCEPT_A_CHILD", "OMOP_CONCEPT_A_CHILD2"}
        elif parent_code == "OMOP_CONCEPT_B":
            return {"OMOP_CONCEPT_B_CHILD"}
        elif parent_code == "OMOP_CONCEPT_A_CHILD":
            return {"OMOP_CONCEPT_A_CHILD_CHILD"}
        else:
            return set()


class DummyLabeler_OMOPConcept(OMOPConceptCodeLabeler):
    original_omop_concept_codes = [
        "OMOP_CONCEPT_A",
        "OMOP_CONCEPT_B",
    ]


def test_omop_concept_code_labeler(tmp_path: pathlib.Path):
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=10))
    ontology = DummyOntology_OMOPConcept()
    labeler = DummyLabeler_OMOPConcept(ontology, time_horizon, prediction_codes=["1", "2"])  # type: ignore
    assert set(labeler.outcome_codes) == {
        "OMOP_CONCEPT_A_CHILD_CHILD",
        "OMOP_CONCEPT_B",
        "OMOP_CONCEPT_B_CHILD",
        "OMOP_CONCEPT_A_CHILD2",
        "OMOP_CONCEPT_A",
        "OMOP_CONCEPT_A_CHILD",
    }
    assert labeler.prediction_codes == ["1", "2"]
    assert labeler.get_time_horizon() == time_horizon


#############################################
#############################################
#
# Specific instances of CodeLabeler
#
#############################################
#############################################


#############################################
# MortalityCodeLabeler
#############################################


class DummyOntology_Mortality(DummyOntology_Base):
    def get_children(self, parent_code: str) -> Set[str]:
        return set()


def test_MortalityCodeLabeler() -> None:
    """Create a MortalityCodeLabeler for codes 3 and 6"""
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))
    events_with_labels: EventsWithLabels = [
        (((1995, 1, 3), 0, 34.5), False),
        (((2000, 1, 1), 1, "test_value"), True),
        (((2000, 1, 5), 2, 1), True),
        (((2000, 6, 5), "SNOMED/419620001", True), "skip"),
        (((2005, 2, 5), 2, None), False),
        (((2005, 7, 5), 2, None), False),
        (((2010, 10, 5), 1, None), False),
        (((2015, 2, 5, 0), 2, None), False),
        (((2015, 7, 5, 0), 0, None), True),
        (((2015, 11, 5, 10, 10), 2, None), True),
        (((2015, 11, 15, 11), "SNOMED/419620001", None), "skip"),
        (((2020, 1, 1), 2, None), "out of range"),
        (((2020, 3, 1, 10, 10, 10), 2, None), "out of range"),
    ]

    ontology = DummyOntology_Mortality()

    # Run labeler
    labeler = MortalityCodeLabeler(ontology, time_horizon)  # type: ignore

    run_test_for_labeler(labeler, events_with_labels, help_text="MortalityLabeler")


#############################################
# LupusCodeLabeler
#############################################


class DummyOntology_Lupus(DummyOntology_Base):
    def get_children(self, parent_code: str) -> Set[str]:
        if parent_code == "SNOMED/55464009":
            return {"SNOMED_55464009", "Lupus_child_seven", "Lupus_child_nine", "Lupus_child_ten"}
        else:
            return set()


def test_LupusCodeLabeler() -> None:
    """Create a LupusCodeLabeler for codes 3 and 6"""
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=180))
    events_with_labels: EventsWithLabels = [
        (((1995, 1, 3), 0, 34.5), False),
        (((2000, 1, 1), 1, "test_value"), True),
        (((2000, 1, 5), 2, 1), True),
        (((2000, 5, 5), "SNOMED/201436003", None), "skip"),
        (((2005, 2, 5), 2, None), False),
        (((2005, 7, 5), 2, None), False),
        (((2010, 10, 5), 1, None), True),
        (((2010, 10, 8), "Lupus_child_seven", None), "skip"),
        (((2015, 2, 5, 0), 2, None), False),
        (((2015, 7, 5, 0), 0, None), True),
        (((2015, 11, 5, 10, 10), 2, None), True),
        (((2015, 11, 15, 11), "SNOMED/55464009", None), "skip"),
        (((2020, 1, 1), "Lupus_child_ten", None), "skip"),
        (((2020, 3, 1, 10, 10, 10), 2, None), "out of range"),
    ]

    ontology = DummyOntology_Lupus()
    labeler = LupusCodeLabeler(ontology, time_horizon)  # type: ignore
    # Check that we selected the right codes
    assert set(labeler.outcome_codes) == {
        "SNOMED_55464009",
        "SNOMED/201436003",
        "Lupus_child_nine",
        "SNOMED/55464009",
        "Lupus_child_ten",
        "Lupus_child_seven",
    }

    run_test_for_labeler(labeler, events_with_labels, help_text="LupusCodeLabeler")


#############################################
#############################################
#
# Specific instances of OMOPConceptCodeLabeler
#
#############################################
#############################################


class DummyOntology_OMOPConcept_Specific(DummyOntology_Base):
    def __init__(self, new_codes: List[str]):
        self.new_codes = new_codes + ["", ""]

    def get_children(self, parent_code: str) -> Set[str]:
        if parent_code == "child_1":
            return {"child_1_1"}
        elif parent_code == self.new_codes[0]:
            return {"child_1"}
        elif parent_code == self.new_codes[1]:
            return {"child_2"}
        return set()


def _assert_labvalue_constructor_correct(
    labeler: OMOPConceptCodeLabeler,
    time_horizon: TimeHorizon,
    outcome_codes: set,
):
    assert set(labeler.outcome_codes) == outcome_codes
    assert labeler.prediction_codes == ["1", "2"]
    assert labeler.get_time_horizon() == time_horizon


def _create_specific_labvalue_labeler(LabelerClass, outcome_codes: set):
    time_horizon = TimeHorizon(datetime.timedelta(days=0), datetime.timedelta(days=10))
    ontology = DummyOntology_OMOPConcept_Specific(LabelerClass.original_omop_concept_codes)
    labeler = LabelerClass(ontology, time_horizon, prediction_codes=["1", "2"])  # type: ignore
    _assert_labvalue_constructor_correct(labeler, time_horizon, outcome_codes)
    return labeler


def test_thrombocytopenia(tmp_path: pathlib.Path):
    outcome_codes = {"child_1_1", "child_2", "SNOMED/89627008", "child_1", "SNOMED/267447008"}
    _create_specific_labvalue_labeler(ThrombocytopeniaCodeLabeler, outcome_codes)


def test_hyperkalemia(tmp_path: pathlib.Path):
    outcome_codes = {"child_1", "SNOMED/14140009", "child_1_1"}
    _create_specific_labvalue_labeler(HyperkalemiaCodeLabeler, outcome_codes)


def test_hypoglycemia(tmp_path: pathlib.Path):
    outcome_codes = {
        "SNOMED/120731000119103",
        "child_2",
        "child_1",
        "SNOMED/52767006",
        "SNOMED/719216001",
        "SNOMED/302866003",
        "SNOMED/267384006",
        "SNOMED/421725003",
        "SNOMED/237637005",
        "SNOMED/237633009",
        "SNOMED/190448007",
        "child_1_1",
        "SNOMED/421437000",
        "SNOMED/230796005",
        "SNOMED/84371000119108",
    }
    _create_specific_labvalue_labeler(HypoglycemiaCodeLabeler, outcome_codes)


def test_hyponatremia(tmp_path: pathlib.Path):
    outcome_codes = {"child_2", "child_1", "SNOMED/89627008", "SNOMED/267447008", "child_1_1"}
    _create_specific_labvalue_labeler(HyponatremiaCodeLabeler, outcome_codes)


def test_anemia(tmp_path: pathlib.Path):
    outcome_codes = {
        "child_1",
        "SNOMED/713496008",
        "SNOMED/691411000119101",
        "SNOMED/691401000119104",
        "SNOMED/767657005",
        "child_2",
        "SNOMED/111570005",
        "SNOMED/271737000",
        "SNOMED/713349004",
        "child_1_1",
    }
    _create_specific_labvalue_labeler(AnemiaCodeLabeler, outcome_codes)


def test_neutropenia(tmp_path: pathlib.Path):
    outcome_codes = {"child_1", "SNOMED/165517008", "child_1_1"}
    _create_specific_labvalue_labeler(NeutropeniaCodeLabeler, outcome_codes)


def test_aki(tmp_path: pathlib.Path):
    outcome_codes = {"child_2", "child_1_1", "child_1", "SNOMED/298015003", "SNOMED/14669001", "SNOMED/35455006"}
    _create_specific_labvalue_labeler(AKICodeLabeler, outcome_codes)



"""Transforms that are unique to STARR OMOP."""

import datetime
from typing import Dict, List, Tuple

import meds


def _move_date_to_end(
    d: datetime.datetime,
) -> datetime.datetime:
    if d.time() == datetime.time.min:
        return d + datetime.timedelta(days=1) - datetime.timedelta(minutes=1)
    else:
        return d


def move_visit_start_to_first_event_start(patient: meds.Patient) -> meds.Patient:
    """Assign visit start times to equal start time of first event in visit

    This function assigns the start time associated with each visit to be
    the start time of the first event that (1) is associated with the visit
    (i.e., shares the same visit ID as the visit event), (2) is a non-visit
    event, and (3) occurs on the same day as the visit event. If the visit
    has no non-visit events or all events associated with the visit have
    the same start time as the visit event (e.g., events with a start time
    of midnight such as billing codes, assuming visit events also have a
    midnight start time) then the visit start time remains unchanged.
    Events that occur on days prior to the visit do not affect the visit
    start time.

    Note that not all visit start times are set to 12:00 AM in the raw data.
    STARR-OMOP currently uses the first available value out of (1) hospital
    admission time, (2) effective date datetime, and (3) effective date, in
    that order. In the OMOP DEID from 12/20/2022 about 10% of visits have
    a time that is not  '00:00:00'.
    """
    first_event_starts: Dict[int, datetime.datetime] = {}
    visit_starts: Dict[int, datetime.datetime] = {}

    # Find the stated start time for each visit
    for event in patient["events"]:
        for measurement in event["measurements"]:
            if measurement["metadata"]["table"] == "visit_occurrence":
                if (
                    measurement["metadata"]["visit_id"] in visit_starts
                    and visit_starts[measurement["metadata"]["visit_id"]] != event["time"]
                ):
                    raise RuntimeError(
                        f"Multiple visit events with visit ID {measurement['metadata']['visit_id']} "
                        + f" for patient ID {patient['patient_id']}"
                    )
                visit_starts[measurement["metadata"]["visit_id"]] = event["time"]

    # Find the minimum start time over all non-visit events associated with each visit
    for event in patient["events"]:
        for measurement in event["measurements"]:
            if measurement["metadata"]["visit_id"] is not None:
                # Only trigger for non-visit events with start time after associated visit start
                # Note: ignores non-visit events starting same time as visit (i.e., at midnight)
                if (
                    measurement["metadata"]["visit_id"] in visit_starts
                    and event["time"] > visit_starts[measurement["metadata"]["visit_id"]]
                ):
                    first_event_starts[measurement["metadata"]["visit_id"]] = min(
                        event["time"],
                        first_event_starts.get(measurement["metadata"]["visit_id"], event["time"]),
                    )

    # Assign visit start times to be same as first non-visit event with same visit ID
    new_events: List[meds.Event] = []
    for event in patient["events"]:
        new_measurements = []
        for measurement in event["measurements"]:
            if measurement["metadata"]["table"] == "visit_occurrence":
                # Triggers if there is a non-visit event associated with the visit ID that has
                # start time strictly after the recorded visit start
                if measurement["metadata"]["visit_id"] in first_event_starts:
                    new_events.append(
                        {"time": first_event_starts[measurement["metadata"]["visit_id"]], "measurements": [measurement]}
                    )
                else:
                    new_measurements.append(measurement)

                if measurement["metadata"].get("end") is not None:
                    # Reset the visit end to be ≥ the visit start
                    measurement["metadata"]["end"] = max(event["time"], measurement["metadata"]["end"])
            else:
                new_measurements.append(measurement)

        if len(new_measurements) > 0:
            new_events.append({"time": event["time"], "measurements": new_measurements})

    patient["events"] = new_events
    patient["events"].sort(key=lambda a: a["time"])

    return patient


def move_to_day_end(patient: meds.Patient) -> meds.Patient:
    """We assume that everything coded at midnight should actually be moved to the end of the day."""
    for event in patient["events"]:
        event["time"] = _move_date_to_end(event["time"])
        for measurement in event["measurements"]:
            if measurement["metadata"].get("end") is not None:
                measurement["metadata"]["end"] = _move_date_to_end(measurement["metadata"]["end"])
                measurement["metadata"]["end"] = max(measurement["metadata"]["end"], event["time"])

    patient["events"].sort(key=lambda a: a["time"])

    return patient


def switch_to_icd10cm(patient: meds.Patient) -> meds.Patient:
    """Switch from ICD10 to ICD10CM."""
    for event in patient["events"]:
        for measurement in event["measurements"]:
            if measurement["code"].startswith("ICD10/"):
                measurement["code"] = measurement["code"].replace("ICD10/", "ICD10CM/", 1)

    return patient


def move_pre_birth(patient: meds.Patient) -> meds.Patient:
    """Move all events to after the birth of a patient."""
    birth_date = None
    for event in patient["events"]:
        for measurement in event["measurements"]:
            if measurement["code"] == meds.birth_code:
                birth_date = event["time"]

    assert birth_date is not None

    new_events = []
    for event in patient["events"]:
        if event["time"] < birth_date:
            delta = birth_date - event["time"]
            if delta > datetime.timedelta(days=30):
                continue

            event["time"] = birth_date

            for measurement in event["measurements"]:
                if measurement["metadata"].get("end") is not None and measurement["metadata"]["end"] < birth_date:
                    measurement["metadata"]["end"] = birth_date

        new_events.append(event)

    patient["events"] = new_events
    patient["events"].sort(key=lambda a: a["time"])

    return patient


def move_billing_codes(patient: meds.Patient) -> meds.Patient:
    """Move billing codes to the end of each visit.

    One issue with our OMOP extract is that billing codes are incorrectly assigned at the start of the visit.
    This class fixes that by assigning them to the end of the visit.
    """
    end_visits: Dict[int, datetime.datetime] = {}  # Map from visit ID to visit end time
    lowest_visit: Dict[Tuple[datetime.datetime, str], int] = {}  # Map from code/start time pairs to visit ID

    # List of billing code tables based on the original Clarity queries used to form STRIDE
    billing_codes = [
        "pat_enc_dx",
        "hsp_acct_dx_list",
        "arpb_transactions",
    ]

    all_billing_codes = {(prefix + "_" + billing_code) for billing_code in billing_codes for prefix in ["shc", "lpch"]}

    for event in patient["events"]:
        for measurement in event["measurements"]:
            # For events that share the same code/start time, we find the lowest visit ID
            if (
                measurement["metadata"].get("clarity_table") in all_billing_codes
                and measurement["metadata"]["visit_id"] is not None
            ):
                key = (event["time"], measurement["code"])
                if key not in lowest_visit:
                    lowest_visit[key] = measurement["metadata"]["visit_id"]
                else:
                    lowest_visit[key] = min(lowest_visit[key], measurement["metadata"]["visit_id"])

            if measurement["metadata"].get("clarity_table") in ("lpch_pat_enc", "shc_pat_enc"):
                if measurement["metadata"].get("end") is not None:
                    if measurement["metadata"]["visit_id"] is None:
                        # Every event with an end time should have a visit ID associated with it
                        raise RuntimeError(f"Expected visit id for visit? {patient['patient_id']} {event}")
                    if (
                        end_visits.get(measurement["metadata"]["visit_id"], measurement["metadata"]["end"])
                        != measurement["metadata"]["end"]
                    ):
                        # Every event associated with a visit should have an end time that matches the visit end time
                        # Also the end times of all events associated with a visit should have the same end time
                        raise RuntimeError(
                            f"Multiple end visits? {end_visits.get(measurement['metadata']['visit_id'])} {event}"
                        )
                    end_visits[measurement["metadata"]["visit_id"]] = measurement["metadata"]["end"]

    new_events: List[meds.Event] = []
    for event in patient["events"]:
        new_measurements: List[meds.Measurement] = []
        for measurement in event["measurements"]:
            if measurement["metadata"].get("clarity_table") in all_billing_codes:
                key = (event["time"], measurement["code"])
                if measurement["metadata"]["visit_id"] != lowest_visit.get(key, None):
                    # Drop this event as we already have it, just with a different visit_id?
                    # We only keep the copy of the event associated with the lowest visit id
                    # (Lowest visit id is arbitrary, no explicit connection to time)
                    continue

                if measurement["metadata"]["visit_id"] is None:
                    # This is a bad code (it has no associated visit_id), but
                    # we would rather keep it than get rid of it
                    new_measurements.append(measurement)
                    continue

                end_visit = end_visits.get(measurement["metadata"]["visit_id"])
                if end_visit is None:
                    raise RuntimeError(f"Expected visit end for code {patient['patient_id']} {event} {patient}")

                # The end time for an event should be no later than its associated visit end time
                if measurement["metadata"].get("end") is not None:
                    measurement["metadata"]["end"] = max(measurement["metadata"]["end"], end_visit)

                # The start time for an event should be no later than its associated visit end time
                new_events.append({"time": max(event["time"], end_visit), "measurements": [measurement]})
            else:
                new_measurements.append(measurement)

        if len(new_measurements) > 0:
            new_events.append({"time": event["time"], "measurements": new_measurements})

    patient["events"] = new_events
    patient["events"].sort(key=lambda a: a["time"])

    return patient



import datetime
from typing import Set

import femr.models.tasks


class DummyOntology:
    def get_all_parents(self, code: str) -> Set[str]:
        if code == "2":
            return {"2", "2_parent"}
        else:
            return {code}


def test_calculator():
    patient = {
        "patient_id": 100,
        "events": [
            {
                "time": datetime.datetime(1990, 1, 10),
                "measurements": [
                    {"code": "1"},
                ],
            },
            {
                "time": datetime.datetime(1990, 1, 20),
                "measurements": [
                    {"code": "2"},
                ],
            },
            {
                "time": datetime.datetime(1990, 1, 25),
                "measurements": [
                    {"code": "3"},
                ],
            },
            {
                "time": datetime.datetime(1990, 1, 25),
                "measurements": [
                    {"code": "1"},
                ],
            },
        ],
    }

    calculator = femr.models.tasks.SurvivalCalculator(DummyOntology(), patient)

    assert calculator.get_future_events_for_time(datetime.datetime(1990, 1, 1)) == (
        datetime.timedelta(days=24),
        {
            "1": datetime.timedelta(days=9),
            "2": datetime.timedelta(days=19),
            "2_parent": datetime.timedelta(days=19),
            "3": datetime.timedelta(days=24),
        },
    )
    assert calculator.get_future_events_for_time(datetime.datetime(1990, 1, 10)) == (
        datetime.timedelta(days=15),
        {
            "1": datetime.timedelta(days=15),
            "2": datetime.timedelta(days=10),
            "2_parent": datetime.timedelta(days=10),
            "3": datetime.timedelta(days=15),
        },
    )
    assert calculator.get_future_events_for_time(datetime.datetime(1990, 1, 20)) == (
        datetime.timedelta(days=5),
        {"1": datetime.timedelta(days=5), "3": datetime.timedelta(days=5)},
    )



from __future__ import annotations

import csv
import dataclasses
import hashlib
import struct
from typing import List

import datasets

import femr.index


@dataclasses.dataclass
class PatientSplit:
    train_patient_ids: List[int]
    test_patient_ids: List[int]

    def save_to_csv(self, fname: str):
        with open(fname, "w") as f:
            writer = csv.DictWriter(f, ("patient_id", "split_name"))
            writer.writeheader()
            for train in self.train_patient_ids:
                writer.writerow({"patient_id": train, "split_name": "train"})
            for test in self.test_patient_ids:
                writer.writerow({"patient_id": test, "split_name": "test"})

    @classmethod
    def load_from_csv(cls, fname: str):
        train_patient_ids: List[int] = []
        test_patient_ids: List[int] = []
        with open(fname, "r") as f:
            for row in csv.DictReader(f):
                if row["split_name"] == "train":
                    train_patient_ids.append(int(row["patient_id"]))
                else:
                    test_patient_ids.append(int(row["patient_id"]))

        return PatientSplit(train_patient_ids=train_patient_ids, test_patient_ids=test_patient_ids)

    def split_dataset(self, dataset: datasets.Dataset, index: femr.index.PatientIndex) -> datasets.DatasetDict:
        train_indices = [index.get_index(patient_id) for patient_id in self.train_patient_ids]
        test_indices = [index.get_index(patient_id) for patient_id in self.test_patient_ids]
        return datasets.DatasetDict(
            {
                "train": dataset.select(train_indices),
                "test": dataset.select(test_indices),
            }
        )


def generate_hash_split(patient_ids: List[int], seed: int, frac_test: float = 0.15) -> PatientSplit:
    train_patient_ids = []
    test_patient_ids = []

    for patient_id in patient_ids:
        # Convert the integer to bytes
        value_bytes = struct.pack(">q", seed) + struct.pack(">q", patient_id)

        # Calculate SHA-256 hash
        sha256_hash = hashlib.sha256(value_bytes).hexdigest()

        # Convert the hexadecimal hash to an integer
        hash_int = int(sha256_hash, 16)

        # Take the modulus
        result = hash_int % (2**16)
        if result <= frac_test * (2**16):
            test_patient_ids.append(patient_id)
        else:
            train_patient_ids.append(patient_id)

    return PatientSplit(train_patient_ids=train_patient_ids, test_patient_ids=test_patient_ids)



"""Core featurizer functionality, shared across Featurizers."""
from __future__ import annotations

import collections
import datetime
import functools
from abc import ABC, abstractmethod
from typing import Any, List, Mapping, NamedTuple, TypeVar

import datasets
import meds
import numpy as np
import scipy.sparse

import femr.index
import femr.ontology


class ColumnValue(NamedTuple):
    """A value for a particular column
    `column` is the index for the column
    `value` is the value for that column. Values must be numeric
    """

    column: int
    value: float | int


def _preprocess_map_func(
    batch, *, label_map: Mapping[int, List[meds.Label]], featurizers: List[Featurizer]
) -> List[List[Any]]:
    patients: List[meds.Patient] = [
        {"patient_id": patient_id, "events": events} for patient_id, events in zip(batch["patient_id"], batch["events"])
    ]

    result = []
    for featurizer in featurizers:
        result.append([featurizer.generate_preprocess_data(patients, label_map)])

    return result


def _preprocess_agg_func(first: List[List[Any]], second: List[List[Any]]) -> List[List[Any]]:
    for a, b in zip(first, second):
        a.extend(b)

    return first


def _features_map_func(
    batch, *, label_map: Mapping[int, List[meds.Label]], featurizers: List[Featurizer]
) -> Mapping[str, Any]:
    # Construct CSR sparse matrix
    #   non-zero entries in sparse matrix
    data: List[Any] = []
    #   maps each element in `data`` to its column in the sparse matrix
    indices: List[int] = []
    #   maps each element in `data` and `indices` to the rows of the sparse matrix
    indptr: List[int] = []

    patient_ids: List[int] = []
    feature_times: List[datetime.datetime] = []

    for patient_id, events in zip(batch["patient_id"], batch["events"]):
        patient: meds.Patient = {"patient_id": patient_id, "events": events}
        labels = label_map[patient_id]

        assert len(labels) != 0, "Must have at least one label per patient processed"

        for label in labels:
            patient_ids.append(patient_id)
            feature_times.append(label["prediction_time"])

        # For each Featurizer, apply it to this Patient...
        features_per_label: List[List[List[ColumnValue]]] = [[] for _ in range(len(labels))]
        for featurizer in featurizers:
            features: List[List[ColumnValue]] = featurizer.featurize(patient, labels)
            assert len(features) == len(labels), (
                f"The featurizer `{featurizer}` didn't generate a set of features for "
                f"every label for patient {patient_id} ({len(features)} != {len(labels)})"
            )
            for a, b in zip(features_per_label, features):
                a.append(b)

        for features in features_per_label:
            indptr.append(len(indices))

            # Keep track of starting column for each successive featurizer as we
            # combine their features into one large matrix
            column_offset: int = 0
            for featurizer, feature_columns in zip(featurizers, features):
                for column, value in feature_columns:
                    assert 0 <= column < featurizer.get_num_columns(), (
                        f"The featurizer {featurizer} provided an out of bounds column for "
                        f"{column} on patient {patient_id} ({column} must be between 0 and "
                        f"{featurizer.get_num_columns()})"
                    )
                    indices.append(column_offset + column)
                    data.append(value)

                # Record what the starting column should be for the next featurizer
                column_offset += featurizer.get_num_columns()

    # Need one last `indptr` for end of last row in CSR sparse matrix
    indptr.append(len(indices))

    # n_rows = number of Labels across all Patients
    total_rows: int = len(indptr) - 1
    # n_cols = sum of number of columns output by each Featurizer
    total_columns: int = sum(x.get_num_columns() for x in featurizers)

    # Explanation of CSR Matrix: https://stackoverflow.com/questions/52299420/scipy-csr-matrix-understand-indptr
    np_data: np.ndarray = np.array(data, dtype=np.float32)
    np_indices: np.ndarray = np.array(indices, dtype=np.int64)
    np_indptr: np.ndarray = np.array(indptr, dtype=np.int64)

    assert (
        np_indptr.shape[0] == total_rows + 1
    ), f"`indptr` length should be equal to '{total_rows + 1}', but instead is '{np_indptr.shape[0]}"
    assert (
        np_data.shape == np_indices.shape
    ), f"`data` should have equal shape as `indices`, but instead have {np_data.shape} != {np_indices.shape}"
    data_matrix = scipy.sparse.csr_matrix((np_data, np_indices, np_indptr), shape=(total_rows, total_columns))

    np_patient_ids: np.ndarray = np.array(patient_ids, dtype=np.int64)
    np_feature_times: np.ndarray = np.array(feature_times, dtype="datetime64[us]")

    return {"patient_ids": [np_patient_ids], "feature_times": [np_feature_times], "features": [data_matrix]}


def _features_agg_func(first_result: Any, second_result: Any) -> Any:
    for k in first_result:
        first_result[k].extend(second_result[k])

    return first_result


class Featurizer(ABC):
    """A Featurizer takes a Patient and a list of Labels, then returns a row for each timepoint.
    Featurizers must be preprocessed before they are used to compute normalization statistics.
    A sparse representation named ColumnValue is used to represent the values returned by a Featurizer.
    """

    def generate_preprocess_data(self, patients: List[meds.Patient], label_map: Mapping[int, List[meds.Label]]) -> Any:
        """
        Some featurizers need to do some preprocessing in order to prepare for featurization.
        This function performs that preprocessing on the given patients and labels, and returns some state.
        That state is concatinated across the entire database,
            and then passed to encorperate_preprocessed_data.

        Note that this function shouldn't mutate the Featurizer as it will be sharded.
        """
        pass

    def encorperate_prepreprocessed_data(self, data_elements: List[Any]) -> None:
        """
        This encorperates data generated from generate_preprocess_data across the entire dataset into the featurizer

        This should mutate the Featurizer.
        """
        pass

    @abstractmethod
    def get_num_columns(self) -> int:
        """Return the number of columns that this featurizer creates."""
        pass

    @abstractmethod
    def featurize(
        self,
        patient: meds.Patient,
        labels: List[meds.Label],
    ) -> List[List[ColumnValue]]:
        """Featurize the patient such that each label in `labels` has an associated list of features.

        Example:
            return [
                [ ColumnValue(0, 10), ColumnValue(3, 12), ], # features for label 0
                [ ColumnValue(1, 0) ], # features for label 1
                [ ColumnValue(2, 2), ColumnValue(1, 3), ColumnValue(10, 3), ], # features for label 2
                ...
                [ ColumnValue(8, 1.3), ColumnValue(9, 5), ], # features for label n
            ]

        Where each ColumnValue is of the form: (idx of column for this feature, feature value).

        Thus, the List[List[ColumnValue]] represents a 2D sparse matrix, where each row is a distinct
        Label and each (sparse) column is a feature

        Args:
            patient (Patient): A patient to featurize.
            labels (List[Label]): We will generate features for each Label in `labels`.
            ontology (Optional[Ontology]): Ontology for Event codes.

        Returns:
             List[List[ColumnValue]]: A list of 'features' (where 'features' is a list itself) for
             each Label.
                The length of this list of lists == length of `labels`
                    [idx] = corresponds to the Label at `labels[idx]`
                    [value] = List of :class:`ColumnValues<ColumnValue>` which contain the features
                    for this label
        """
        pass

    def get_column_name(self, column_idx: int) -> str:
        """Enable the user to get the name of a column by its index

        Args:
            column_idx (int): The index of the column
        """
        return "no name"

    def is_needs_preprocessing(self) -> bool:
        """Return TRUE if you must run `preprocess()`. If FALSE, then `preprocess()`
        should do nothing.
        """
        return False


FeaturizerType = TypeVar("FeaturizerType", bound=Featurizer)


class FeaturizerList:
    """
    FeaturizerList consists of a list of Featurizers that will be used to (sequentially)
    featurize Patients based on their Labels.

    It enables preprocessing of featurizers, featurization, and column name extraction.
    """

    def __init__(self, featurizers: List[Featurizer]):
        """Create a `FeaturizerList` from a sequence of featurizers.

        Args:
            featurizers (List[Featurizer]): The featurizers to use for featurizing patients.
        """
        self.featurizers: List[Featurizer] = featurizers

    def preprocess_featurizers(
        self,
        dataset: datasets.Dataset,
        index: femr.index.PatientIndex,
        labels: List[meds.Label],
        num_proc: int = 1,
    ) -> None:
        """Preprocess `self.featurizers` on the provided set of labels."""

        # Check if any featurizers need preprocessing. If not, return early.
        any_needs_preprocessing: bool = any(featurizer.is_needs_preprocessing() for featurizer in self.featurizers)
        if not any_needs_preprocessing:
            return

        label_map = collections.defaultdict(list)

        for label in labels:
            label_map[label["patient_id"]].append(label)
        # Split patients across multiple threads
        patient_ids: List[int] = sorted(list({label["patient_id"] for label in labels}))

        dataset = index.filter_dataset(dataset, patient_ids)

        # Preprocess in parallel
        featurize_stats = femr.hf_utils.aggregate_over_dataset(
            dataset,
            functools.partial(_preprocess_map_func, label_map=label_map, featurizers=self.featurizers),
            _preprocess_agg_func,
            batch_size=1_000,
            num_proc=num_proc,
        )

        # Aggregate featurizers
        for featurizer, featurizer_stat in zip(self.featurizers, featurize_stats):
            # Merge all featurizers of the same class as `featurizer`
            featurizer.encorperate_prepreprocessed_data(featurizer_stat)

    def featurize(
        self,
        dataset: datasets.Dataset,
        index: femr.index.PatientIndex,
        labels: List[meds.Label],
        num_proc: int = 1,
    ) -> Mapping[str, np.ndarray]:
        """
        Apply a list of Featurizers (in sequence) to obtain a feature matrix for each Label for each patient.

        Args:
            database_path (str): Path to `PatientDatabase` on disk

        Returns:
            This returns a tuple (data_matrix, labels, patient_ids, labeling_time).
                data_matrix is a sparse matrix of all the features of all the featurizers.
                label_pids is a list of the patient ids for each row.
                label_values is a list of boolean values representing the labels for each row in the matrix.
                labeling_time is a list of labeling/prediction time for each row.
        """
        label_map = collections.defaultdict(list)

        for label in labels:
            label_map[label["patient_id"]].append(label)
        # Split patients across multiple threads
        patient_ids: List[int] = sorted(list({label["patient_id"] for label in labels}))

        dataset = index.filter_dataset(dataset, patient_ids)

        features = femr.hf_utils.aggregate_over_dataset(
            dataset,
            functools.partial(_features_map_func, label_map=label_map, featurizers=self.featurizers),
            _features_agg_func,
            batch_size=1_000,
            num_proc=num_proc,
        )

        result = {k: np.concatenate(features[k]) for k in ("patient_ids", "feature_times")}

        result["features"] = scipy.sparse.vstack(features["features"])

        return result

    def get_column_name(self, column_idx: int) -> str:
        column_offset: int = 0
        for featurizer in self.featurizers:
            if column_offset <= column_idx < (column_offset + featurizer.get_num_columns()):
                return f"Featurizer {featurizer}, {featurizer.get_column_name(column_idx - column_offset)}"
            column_offset += featurizer.get_num_columns()
        raise IndexError(f"Column index '{column_idx}' out of bounds for this FeaturizerList")


def join_labels(features: Mapping[str, np.array], labels: List[meds.Label]) -> Mapping[str, np.array]:
    labels = list(labels)
    labels.sort(key=lambda a: (a["patient_id"], a["prediction_time"]))

    label_index = 0

    indices = []
    label_values = []

    order = np.lexsort((features["feature_times"], features["patient_ids"]))

    for i, patient_id, feature_time in zip(order, features["patient_ids"][order], features["feature_times"][order]):
        if label_index == len(labels):
            break

        assert patient_id <= labels[label_index]["patient_id"], f"Missing features for label {labels[label_index]}"
        if patient_id < labels[label_index]["patient_id"]:
            continue

        assert (
            feature_time <= labels[label_index]["prediction_time"]
        ), f"Missing features for label {labels[label_index]}"
        if feature_time < labels[label_index]["prediction_time"]:
            continue

        indices.append(i)
        label_values.append(labels[label_index]["boolean_value"])
        label_index += 1

    return {
        "boolean_values": np.array(label_values),
        "patient_ids": features["patient_ids"][indices],
        "times": features["feature_times"][indices],
        "features": features["features"][indices, :],
    }



from __future__ import annotations

from typing import Any, Dict, Mapping, Optional

import transformers


class FEMRTransformerConfig(transformers.PretrainedConfig):
    def __init__(
        self,
        vocab_size: int = 32768,
        is_hierarchical: bool = False,
        hidden_size: int = 768,
        intermediate_size: int = 3072,
        n_heads: int = 12,
        n_layers: int = 6,
        attention_width: int = 496,
        use_normed_ages: bool = False,
        use_bias: bool = True,
        hidden_act: str = "gelu",
        **kwargs,
    ) -> None:
        """Defined a configuration for a FEMR Transformer.
        
        Arguments:
            vocab_size: The number of tokens in the vocabulary
            is_hierarchical: Whether to use a hierarchical vocabulary. See FEMRTokenizer for more information.
            hidden_size: The internal representation size
            intermediate_size: The size of the FFN in the transformer layers
            n_heads: The number of attention heads
            n_layers: The number of transformer encoder layers
            attention_width: FEMR by default uses a local attention transformer. This defines the width of those attention windows.
            use_normed_ages: Whether or not to provide normalized ages as a feature to the model
            use_bias: Whether or not to use bias terms in the transformer layers. Current research indicates that this should be set to False.
            hidden_act: The type of activation function to use in the transformer.
            """
        super().__init__(**kwargs)

        self.vocab_size = vocab_size
        self.is_hierarchical = is_hierarchical

        self.hidden_size = hidden_size
        self.intermediate_size = intermediate_size
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.attention_width = attention_width

        self.use_normed_ages = use_normed_ages

        self.use_bias = use_bias
        self.hidden_act = hidden_act


class FEMRTaskConfig(transformers.PretrainedConfig):
    def __init__(self, task_type: str = "", task_kwargs: Mapping[str, Any] = {}, **kwargs):
        """A generic FEMR task definition. This holds state used for initalizing a tasks.py class.

        Task.get_task_config returns the task type and kwargs used to initialize this.
        
        Arguments:
            task_type: The name of the task. 
            task_kwargs: Arbitrary arguments used to store state for that task.
        """
        super().__init__(**kwargs)
        self.task_type = task_type
        self.task_kwargs = task_kwargs


class FEMRModelConfig(transformers.PretrainedConfig):
    """A model config is defined as the combination of a transformer config and a task config."""
    def __init__(
        self,
        transformer_config: Optional[Dict[str, Any]] = None,
        task_config: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """A combination of a transformer config and a task config.
        
        It is possible to initialize this with only a transformer config, in which case the model will be configured for inference only.
        """
        super().__init__(**kwargs)
        if transformer_config is None:
            transformer_config = {}
        self.transformer_config = FEMRTransformerConfig(**transformer_config)

        self.task_config: Optional[FEMRTaskConfig]

        if task_config is not None:
            self.task_config = FEMRTaskConfig(**task_config)
        else:
            self.task_config = None

    @classmethod
    def from_transformer_task_configs(
        cls, transformer_config: FEMRTransformerConfig, task_config: FEMRTaskConfig
    ) -> FEMRModelConfig:
        """
        Combine a transformer configuration and task configuration into a model configuration.
        """
        if task_config is not None:
            task_config_dict = task_config.to_dict()
        else:
            task_config_dict = None

        return cls(transformer_config=transformer_config.to_dict(), task_config=task_config_dict)



from __future__ import annotations

import collections
import datetime
import functools
import random
from collections import defaultdict, deque
from typing import Any, Callable, Deque, Dict, Iterable, Iterator, List, Mapping, Optional, Set, Tuple

import meds
import numpy as np

import femr.ontology

from .core import ColumnValue, Featurizer
from .utils import OnlineStatistics


def get_patient_birthdate(patient: meds.Patient) -> datetime.datetime:
    for e in patient["events"]:
        for m in e["measurements"]:
            if m["code"] == meds.birth_code:
                return e["time"]
    raise ValueError("Couldn't find patient birthdate -- Patient has no events")


class AgeFeaturizer(Featurizer):
    """
    Produces the (possibly normalized) age at each label timepoint.
    """

    def __init__(self, is_normalize: bool = True):
        """
        Args:
            is_normalize (bool, optional): If TRUE, then normalize a patient's age at each
            label across their ages at all labels. Defaults to True.
        """
        self.is_normalize: bool = is_normalize
        self.age_statistics: Optional[OnlineStatistics] = None

    def get_num_columns(self) -> int:
        return 1

    def generate_preprocess_data(
        self, patients: List[meds.Patient], label_map: Mapping[int, List[meds.Label]]
    ) -> OnlineStatistics:
        """Save the age of this patient (in years) at each label, to use for normalization."""
        if not self.is_needs_preprocessing():
            return OnlineStatistics()

        age_statistics: OnlineStatistics = OnlineStatistics()

        for patient in patients:
            patient_birth_date: Optional[datetime.datetime] = get_patient_birthdate(patient)
            assert patient_birth_date, "Patients must have a birth date"

            for label in label_map[patient["patient_id"]]:
                age_in_yrs: float = (label["prediction_time"] - patient_birth_date).days / 365
                age_statistics.add(age_in_yrs)

        return age_statistics

    def encorperate_prepreprocessed_data(self, data_elements: List[OnlineStatistics]) -> None:
        self.age_statistics = OnlineStatistics.merge(data_elements)

    def featurize(
        self,
        patient: meds.Patient,
        labels: List[meds.Label],
    ) -> List[List[ColumnValue]]:
        """Return the age of the patient at each label.
        If `is_normalize`, then normalize each label's age across all patient's ages across all their labels."""
        all_columns: List[List[ColumnValue]] = []
        # Outer list is per label
        # Inner list is the list of features for that label

        patient_birth_date: Optional[datetime.datetime] = get_patient_birthdate(patient)
        if not patient_birth_date:
            return all_columns

        for label in labels:
            age_in_yrs: float = (label["prediction_time"] - patient_birth_date).days / 365
            if self.is_normalize:
                assert self.age_statistics is not None
                # age = (age - mean(ages)) / std(ages)
                age_in_yrs = (age_in_yrs - self.age_statistics.mean()) / (self.age_statistics.standard_deviation())
            all_columns.append([ColumnValue(0, age_in_yrs)])

        return all_columns

    def is_needs_preprocessing(self) -> bool:
        return self.is_normalize

    def __repr__(self):
        return (
            f"AgeFeaturizer(is_normalize={self.is_normalize}, count={self.age_statistics.current_count}"
            f" mean={self.age_statistics.mean()}, std={self.age_statistics.standard_deviation()})"
        )


def _reshuffle_count_time_bins(
    time_bins: List[datetime.timedelta],
    codes_per_bin: Dict[int, Deque[Tuple[int, datetime.datetime]]],
    code_counts_per_bin: Dict[int, Dict[int, int]],
    label: meds.Label,
):
    # From closest bin to prediction time -> farthest bin
    for bin_idx, bin_end in enumerate(time_bins):
        while len(codes_per_bin[bin_idx]) > 0:
            # Get the least recently added event (i.e. farthest back in patient's timeline
            # from the currently processed label)
            oldest_event_code, oldest_event_start = codes_per_bin[bin_idx][0]

            if (label["prediction_time"] - oldest_event_start) <= bin_end:
                # The oldest event that we're tracking is still within the closest (i.e. smallest distance)
                # bin to our label's prediction time, so all events will be within this bin,
                # so we don't have to worry about shifting events into farther bins (as the code
                # in the `else` clause does)
                break
            else:
                # Goal: Readjust codes so that they fall under the proper time bin
                # Move (oldest_event_code, oldest_event_start) from entry @ `bin_idx`
                # to entry @ `bin_idx + 1`.
                # Basically, move this code from the bin that is closer to the prediction time (`bin_idx`)
                # to a bin that is further away from the prediction time (`bin_idx + 1`)
                codes_per_bin[bin_idx + 1].append(codes_per_bin[bin_idx].popleft())

                # Remove oldest_event_code from current (closer to prediction time) bin `bin_idx`
                code_counts_per_bin[bin_idx][oldest_event_code] -= 1
                # Add oldest_event_code to the (farther from prediction time) bin `bin_idx + 11
                code_counts_per_bin[bin_idx + 1][oldest_event_code] += 1

                # Clear out ColumnValues with a value of 0 to preserve sparsity of matrix
                if code_counts_per_bin[bin_idx][oldest_event_code] == 0:
                    del code_counts_per_bin[bin_idx][oldest_event_code]


class ReservoirSampler:
    def __init__(self, k, rng_seed):
        self.k = k
        self.total = 0
        self.values = []
        self.rng = random.Random(rng_seed)

    def add(self, value):
        if len(self.values) < self.k:
            self.values.append(value)
        else:
            r = self.rng.randint(0, self.total)
            if r < self.k:
                self.values[r] = value

        self.total += 1


def exclusion_helper(
    measurement: meds.Measurement, fallback_function: Callable[[meds.Measurement], bool], excluded_codes_set: Set[str]
) -> bool:
    if excluded_codes_set is not None:
        if measurement["code"] in excluded_codes_set:
            return True
    if fallback_function is not None:
        return fallback_function(measurement)

    return False


class CountFeaturizer(Featurizer):
    """
    Produces one column per each diagnosis code, procedure code, and prescription code.
    The value in each column is the count of how many times that code appears in the patient record
    before the corresponding label.
    """

    def __init__(
        self,
        ontology: Optional[femr.ontology.Ontology] = None,
        is_ontology_expansion: bool = False,
        excluded_codes: Iterable[str] = [],
        excluded_event_filter: Optional[Callable[[meds.Measurement], bool]] = None,
        time_bins: Optional[List[datetime.timedelta]] = None,
        numeric_value_decile: bool = False,
        string_value_combination: bool = False,
        characters_for_string_values: int = 100,
    ):
        """
        Args:
            is_ontology_expansion (bool, optional): If TRUE, then do ontology expansion when counting codes.

                Example:
                    If `is_ontology_expansion=True` and your ontology is:
                        Code A -> Code B -> Code C
                    Where "->" denotes "is a parent of" relationship (i.e. A is a parent of B, B is a parent of C).
                    Then if we see 2 occurrences of Code "C", we count 2 occurrences of Code "B" and Code "A".

            excluded_codes (List[str], optional): A list of femr codes that we will ignore. Defaults to [].

            time_bins (Optional[List[datetime.timedelta]], optional): Group counts into buckets.
                Starts from the label time, and works backwards according to each successive value in `time_bins`.

                These timedeltas should be positive values, and will be internally converted to negative values

                If last value is `None`, then the last bucket will be from the penultimate value in `time_bins` to the
                    start of the patient's first event.

                Examples:
                    `time_bins = [
                        datetime.timedelta(days=90),
                        datetime.timedelta(days=180)
                    ]`
                        will create the following buckets:
                            [label time, -90 days], [-90 days, -180 days];
                    `time_bins = [
                        datetime.timedelta(days=90),
                        datetime.timedelta(days=180),
                        datetime.timedelta(years=100)
                    ]`
                        will create the following buckets:
                            [label time, -90 days], [-90 days, -180 days], [-180 days, -100 years];]
        """
        self.ontology = ontology
        self.is_ontology_expansion: bool = is_ontology_expansion
        self.excluded_event_filter = functools.partial(
            exclusion_helper, fallback_function=excluded_event_filter, excluded_codes_set=set(excluded_codes)
        )
        self.time_bins: Optional[List[datetime.timedelta]] = time_bins
        self.characters_for_string_values: int = characters_for_string_values

        self.numeric_value_decile = numeric_value_decile
        self.string_value_combination = string_value_combination

        if self.time_bins is not None:
            assert len(set(self.time_bins)) == len(
                self.time_bins
            ), f"You cannot have duplicate values in the `time_bins` argument. You passed in: {self.time_bins}"

        self.finalized = False

    def get_codes(self, code: str) -> Iterator[str]:
        if self.is_ontology_expansion:
            assert self.ontology is not None
            for subcode in self.ontology.get_all_parents(code):
                yield subcode
        else:
            yield code

    def get_columns(self, measurement: meds.Measurement) -> Iterator[int]:
        if measurement.get("text_value"):
            k = (measurement["code"], measurement["text_value"][: self.characters_for_string_values])
            if k in self.code_string_to_column_index:
                yield self.code_string_to_column_index[k]
        elif measurement.get("numeric_value"):
            if measurement["code"] in self.code_value_to_column_index:
                column, quantiles = self.code_value_to_column_index[measurement["code"]]
                for i, (start, end) in enumerate(zip(quantiles, quantiles[1:])):
                    if start <= measurement["numeric_value"] < end:
                        yield i + column
        else:
            for code in self.get_codes(measurement["code"]):
                # If we haven't seen this code before, then add it to our list of included codes
                if code in self.code_to_column_index:
                    yield self.code_to_column_index[code]

    def generate_preprocess_data(self, patients: List[meds.Patient], label_map: Mapping[int, List[meds.Label]]) -> Any:
        """
        Some featurizers need to do some preprocessing in order to prepare for featurization.
        This function performs that preprocessing on the given patients and labels, and returns some intermediate state.
        That state is concatinated across the entire database, and then passed to encorperate_preprocessed_data.

        Note that this function shouldn't mutate the Featurizer as it will be sharded.
        """
        observed_codes: Set[str] = set()
        observed_string_value: Dict[Tuple[str, str], int] = collections.defaultdict(int)
        observed_numeric_value: Dict[str, ReservoirSampler] = collections.defaultdict(
            functools.partial(ReservoirSampler, 10000, 100)
        )

        for patient in patients:
            for event in patient["events"]:
                for measurement in event["measurements"]:
                    # Check for excluded events
                    if self.excluded_event_filter is not None and self.excluded_event_filter(measurement):
                        continue

                    if measurement.get("text_value"):
                        if self.string_value_combination:
                            observed_string_value[
                                (measurement["code"], measurement["text_value"][: self.characters_for_string_values])
                            ] += 1
                    elif measurement.get("numeric_value"):
                        if self.numeric_value_decile:
                            observed_numeric_value[measurement["code"]].add(measurement["numeric_value"])
                    else:
                        for code in self.get_codes(measurement["code"]):
                            # If we haven't seen this code before, then add it to our list of included codes
                            observed_codes.add(code)

        return {
            "observed_codes": observed_codes,
            "observed_string_value": observed_string_value,
            "observed_numeric_value": observed_numeric_value,
        }

    def encorperate_prepreprocessed_data(self, data_elements: List[Any]) -> None:
        """
        This encorperates data generated from generate_preprocess_data across the entire dataset into the featurizer

        This should mutate the Featurizer.
        """

        observed_codes: Set[str] = set()
        observed_string_value: Dict[Tuple[str, str], int] = collections.defaultdict(int)
        observed_numeric_value: Dict[str, ReservoirSampler] = collections.defaultdict(
            functools.partial(ReservoirSampler, 10000, 100)
        )

        for data_element in data_elements:
            observed_codes |= data_element["observed_codes"]
            for k1, v1 in data_element["observed_string_value"].items():
                observed_string_value[k1] += v1
            for k2, v2 in data_element["observed_numeric_value"].items():
                observed_numeric_value[k2].values += v2.values

        self.code_to_column_index = {}
        self.code_string_to_column_index = {}
        self.code_value_to_column_index = {}

        self.num_columns = 0

        for code in sorted(list(observed_codes)):
            self.code_to_column_index[code] = self.num_columns
            self.num_columns += 1

        for (code, val), count in sorted(list(observed_string_value.items())):
            if count > 1:
                self.code_string_to_column_index[(code, val)] = self.num_columns
                self.num_columns += 1

        for code, values in sorted(list(observed_numeric_value.items())):
            quantiles = sorted(list(set(np.quantile(values.values, np.linspace(0, 1, num=11)[1:-1]))))
            quantiles = [float("-inf")] + quantiles + [float("inf")]
            self.code_value_to_column_index[code] = (self.num_columns, quantiles)
            self.num_columns += len(quantiles) - 1

    def get_num_columns(self) -> int:
        if self.time_bins is None:
            return self.num_columns
        else:
            return self.num_columns * len(self.time_bins)

    def featurize(
        self,
        patient: meds.Patient,
        labels: List[meds.Label],
    ) -> List[List[ColumnValue]]:
        all_columns: List[List[ColumnValue]] = []

        if self.time_bins is None:
            # Count the number of times each code appears in the patient's timeline
            # [key] = column idx
            # [value] = count of occurrences of events with that code (up to the label at `label_idx`)
            code_counter: Dict[int, int] = defaultdict(int)

            label_idx = 0
            for event in patient["events"]:
                while event["time"] > labels[label_idx]["prediction_time"]:
                    label_idx += 1
                    # Create all features for label at index `label_idx`
                    all_columns.append([ColumnValue(code, count) for code, count in code_counter.items()])
                    if label_idx >= len(labels):
                        # We've reached the end of the labels for this patient,
                        # so no point in continuing to count events past this point.
                        # Instead, we just return the counts of all events up to this point.
                        return all_columns

                for measurement in event["measurements"]:
                    if self.excluded_event_filter is not None and self.excluded_event_filter(measurement):
                        continue

                    for column_idx in self.get_columns(measurement):
                        code_counter[column_idx] += 1

            # For all labels that occur past the last event, add all
            # events' total counts as these labels' feature values (basically,
            # the featurization of these labels is the count of every single event)
            for _ in labels[label_idx:]:
                all_columns.append([ColumnValue(code, count) for code, count in code_counter.items()])

        else:
            # First, sort time bins in ascending order (i.e. [100 days, 90 days, 1 days] -> [1, 90, 100])
            time_bins: List[datetime.timedelta] = sorted([x for x in self.time_bins if x is not None])

            codes_per_bin: Dict[int, Deque[Tuple[int, datetime.datetime]]] = {
                i: deque() for i in range(len(self.time_bins) + 1)
            }

            code_counts_per_bin: Dict[int, Dict[int, int]] = {
                i: defaultdict(int) for i in range(len(self.time_bins) + 1)
            }

            label_idx = 0
            for event in patient["events"]:
                while event["time"] > labels[label_idx]["prediction_time"]:
                    _reshuffle_count_time_bins(
                        time_bins,
                        codes_per_bin,
                        code_counts_per_bin,
                        labels[label_idx],
                    )
                    label_idx += 1
                    # Create all features for label at index `label_idx`
                    all_columns.append(
                        [
                            ColumnValue(
                                code + i * self.num_columns,
                                count,
                            )
                            for i in range(len(self.time_bins))
                            for code, count in code_counts_per_bin[i].items()
                        ]
                    )

                    if label_idx >= len(labels):
                        # We've reached the end of the labels for this patient,
                        # so no point in continuing to count events past this point.
                        # Instead, we just return the counts of all events up to this point.
                        return all_columns

                for measurement in event["measurements"]:
                    if self.excluded_event_filter is not None and self.excluded_event_filter(measurement):
                        continue

                    for column_idx in self.get_columns(measurement):
                        codes_per_bin[0].append((column_idx, event["time"]))
                        code_counts_per_bin[0][column_idx] += 1

            for label in labels[label_idx:]:
                _reshuffle_count_time_bins(
                    time_bins,
                    codes_per_bin,
                    code_counts_per_bin,
                    label,
                )
                all_columns.append(
                    [
                        ColumnValue(
                            code + i * self.num_columns,
                            count,
                        )
                        for i in range(len(self.time_bins))
                        for code, count in code_counts_per_bin[i].items()
                    ]
                )

        return all_columns

    def is_needs_preprocessing(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"CountFeaturizer(number of included codes={self.num_columns})"

    def get_column_name(self, column_idx: int) -> str:
        def helper(actual_idx):
            for code, idx in self.code_to_column_index.items():
                if idx == actual_idx:
                    return code
            for (code, val), idx in self.code_string_to_column_index.items():
                if idx == actual_idx:
                    return f"{code} {val}"

            for code, (idx, quantiles) in self.code_value_to_column_index.items():
                offset = actual_idx - idx
                if 0 <= offset < len(quantiles) - 1:
                    return f"{code} [{quantiles[offset]}, {quantiles[offset+1]})"

            raise RuntimeError("Could not find name for " + str(actual_idx))

        if self.time_bins is None:
            return helper(column_idx)
        else:
            return helper(column_idx % self.num_columns) + f"_{self.time_bins[column_idx // self.num_columns]}"



import femr.models.processor
import femr.models.tasks
import datetime

from femr_test_tools import create_patients_dataset

class DummyTokenizer:
    def __init__(self, is_hierarchical: bool = False):
        self.is_hierarchical = is_hierarchical
        self.ontology = None
        self.vocab_size = 100

    def start_patient(self):
        pass

    def get_feature_codes(self, time, measurement):
        if measurement['code'] == 'SNOMED/184099003':
            return [1], None
        else:
            return [int(measurement['code'])], None

    def normalize_age(self, age):
        return 0.5

def assert_two_batches_equal_third(batch1, batch2, batch3):
    """This asserts that batch1 + batch2 = batchs3"""
    assert batch3['patient_ids'].tolist() == batch1['patient_ids'].tolist() + batch2['patient_ids'].tolist()

    assert batch3['transformer']['ages'].tolist() == batch1['transformer']['ages'].tolist() + batch2['transformer']['ages'].tolist()
    assert batch3['transformer']['timestamps'].tolist() == batch1['transformer']['timestamps'].tolist() + batch2['transformer']['timestamps'].tolist()


    # Checking the label indices is a bit more involved as we have to map to age/patient id and then check that
    target_label_ages = []
    target_label_patient_ids = []

    for label_index in batch1['transformer']['label_indices'].tolist():
        target_label_ages.append(batch1['transformer']['ages'][label_index])
        target_label_patient_ids.append(batch1['patient_ids'][label_index])

    for label_index in batch2['transformer']['label_indices'].tolist():
        target_label_ages.append(batch2['transformer']['ages'][label_index])
        target_label_patient_ids.append(batch2['patient_ids'][label_index])

    actual_label_ages = []
    actual_label_patient_ids = []

    for label_index in batch3['transformer']['label_indices'].tolist():
        actual_label_ages.append(batch3['transformer']['ages'][label_index])
        actual_label_patient_ids.append(batch3['patient_ids'][label_index])

    assert target_label_ages == actual_label_ages
    assert target_label_patient_ids == actual_label_patient_ids

    if 'tokens' in batch3['transformer']:
        assert batch3['transformer']['tokens'].tolist() == batch1['transformer']['tokens'].tolist() + batch2['transformer']['tokens'].tolist()

def test_two_patients_concat_no_task():
    tokenizer = DummyTokenizer()

    fake_patients = create_patients_dataset(10)

    fake_patient1 = fake_patients[1]
    fake_patient2 = fake_patients[5]

    creator = femr.models.processor.BatchCreator(tokenizer)

    creator.start_batch()
    creator.add_patient(fake_patient1)

    data_for_patient1 = creator.get_batch_data()

    creator.start_batch()
    creator.add_patient(fake_patient2)

    data_for_patient2 = creator.get_batch_data()

    
    creator.start_batch()
    creator.add_patient(fake_patient1)
    creator.add_patient(fake_patient2)

    data_for_patients = creator.get_batch_data()

    assert_two_batches_equal_third(data_for_patient1, data_for_patient2, data_for_patients)

def test_split_patients_concat_no_task():
    tokenizer = DummyTokenizer()

    fake_patients = create_patients_dataset(10)

    fake_patient = fake_patients[1]

    creator = femr.models.processor.BatchCreator(tokenizer)

    creator.start_batch()
    creator.add_patient(fake_patient)

    data_for_patient = creator.get_batch_data()

    length = len(data_for_patient['transformer']['timestamps'])
    
    creator.start_batch()
    creator.add_patient(fake_patient, offset=0, max_length = length // 2)

    data_for_part1 = creator.get_batch_data()

    creator.start_batch()
    creator.add_patient(fake_patient, offset=length // 2, max_length = None)

    data_for_part2 = creator.get_batch_data()

    assert_two_batches_equal_third(data_for_part1, data_for_part2, data_for_patient)

def test_two_patients_concat_task():
    tokenizer = DummyTokenizer()

    fake_patients = create_patients_dataset(10)
    
    task = femr.models.tasks.LabeledPatientTask([
        {'patient_id': 1, 'prediction_time': datetime.datetime(2011, 7, 6)},
        {'patient_id': 1, 'prediction_time': datetime.datetime(2017, 1, 1)},
        
        {'patient_id': 5, 'prediction_time': datetime.datetime(2011, 11, 6)},
        {'patient_id': 5, 'prediction_time': datetime.datetime(2017, 2, 1)}
    ])

    fake_patient1 = fake_patients[1]
    fake_patient2 = fake_patients[5]

    creator = femr.models.processor.BatchCreator(tokenizer, task=task)

    creator.start_batch()
    creator.add_patient(fake_patient1)

    data_for_patient1 = creator.get_batch_data()

    creator.start_batch()
    creator.add_patient(fake_patient2)

    data_for_patient2 = creator.get_batch_data()

    
    creator.start_batch()
    creator.add_patient(fake_patient1)
    creator.add_patient(fake_patient2)

    data_for_patients = creator.get_batch_data()

    assert_two_batches_equal_third(data_for_patient1, data_for_patient2, data_for_patients)

def test_split_patients_concat_task():
    tokenizer = DummyTokenizer()

    fake_patients = create_patients_dataset(10)

    fake_patient = fake_patients[1]

    task = femr.models.tasks.LabeledPatientTask([
        {'patient_id': 1, 'prediction_time': datetime.datetime(2011, 7, 6)},
        {'patient_id': 1, 'prediction_time': datetime.datetime(2017, 1, 1)},
    ])

    creator = femr.models.processor.BatchCreator(tokenizer, task=task)

    creator.start_batch()
    creator.add_patient(fake_patient)

    data_for_patient = creator.get_batch_data()

    length = len(data_for_patient['transformer']['timestamps'])
    
    creator.start_batch()
    creator.add_patient(fake_patient, offset=0, max_length = length // 2)

    data_for_part1 = creator.get_batch_data()

    creator.start_batch()
    creator.add_patient(fake_patient, offset=length // 2, max_length = None)

    data_for_part2 = creator.get_batch_data()

    assert_two_batches_equal_third(data_for_part1, data_for_part2, data_for_patient)


from __future__ import annotations

import bisect
import collections
import datetime
import functools
import math
import os
from typing import Any, Dict, Mapping, Optional, Set, Union, Tuple, List

import meds
import msgpack
import numpy as np
import transformers

import femr.hf_utils
import femr.stat_utils


def train_tokenizer(
    dataset,
    vocab_size: int,
    is_hierarchical: bool = False,
    num_numeric: int = 1000,
    ontology: Optional[femr.ontology.Ontology] = None,
    num_proc: int = 1,
) -> FEMRTokenizer:
    """Train a FEMR tokenizer from the given dataset"""
    statistics = femr.hf_utils.aggregate_over_dataset(
        dataset,
        functools.partial(
            map_statistics, num_patients=len(dataset), is_hierarchical=is_hierarchical, ontology=ontology
        ),
        agg_statistics,
        num_proc=num_proc,
        batch_size=1_000,
    )
    return FEMRTokenizer(
        convert_statistics_to_msgpack(statistics, vocab_size, is_hierarchical, num_numeric, ontology), ontology
    )


def agg_statistics(stats1, stats2):
    stats1["age_stats"].combine(stats2["age_stats"])

    for n in ("code_counts", "text_counts"):
        for k, v in stats2[n].items():
            stats1[n][k] += v

    if stats1.get("numeric_samples"):
        stats1["numeric_samples"].combine(stats2["numeric_samples"])
    if stats1.get("numeric_samples_by_lab"):
        for k, v in stats2["numeric_samples_by_lab"].items():
            stats1["numeric_samples_by_lab"][k].combine(v)

    return stats1


def normalize_unit(unit):
    if unit:
        return unit.lower().replace(" ", "")
    else:
        return None


def map_statistics(
    batch, *, num_patients: int, is_hierarchical: bool, frac_values=0.05, ontology: Optional[femr.ontology.Ontology]
) -> Mapping[str, Any]:
    age_stats = femr.stat_utils.OnlineStatistics()
    code_counts: Dict[str, float] = collections.defaultdict(float)

    numeric_samples: Optional[femr.stat_utils.ReservoirSampler]
    numeric_samples_by_lab: Optional[Dict[str, femr.stat_utils.ReservoirSampler]]
    if is_hierarchical:
        assert ontology is not None
        numeric_samples = femr.stat_utils.ReservoirSampler(10_000)
        numeric_samples_by_lab = None
    else:
        numeric_samples = None
        numeric_samples_by_lab = collections.defaultdict(functools.partial(femr.stat_utils.ReservoirSampler, 1_000))

    text_counts: Dict[Any, float] = collections.defaultdict(float)

    for events in batch["events"]:
        total_events = 0
        for event in events:
            for measurement in event["measurements"]:
                total_events += 1

        if total_events == 0:
            continue

        weight = 1.0 / (num_patients * total_events)
        birth_date = events[0]["time"]
        code_set = set()
        text_set = set()
        pat_numeric_samples = []
        for event in events:
            for measurement in event["measurements"]:
                if event["time"] != birth_date:
                    age_stats.add(weight, (event["time"] - birth_date).total_seconds())
                if not is_hierarchical:
                    assert numeric_samples_by_lab is not None
                    if measurement["numeric_value"] is not None:
                        numeric_samples_by_lab[measurement["code"]].add(measurement["numeric_value"], weight)
                    elif measurement["text_value"] is not None:
                        text_counts[(measurement["code"], measurement["text_value"])] += weight
                    else:
                        code_counts[measurement["code"]] += weight
                else:
                    code_set.add(measurement["code"])

                    if measurement["text_value"] is not None and measurement["text_value"] != "":
                        text_set.add(measurement["text_value"])

                    if measurement.get("metadata") and normalize_unit(measurement["metadata"].get("unit")) is not None:
                        text_set.add(normalize_unit(measurement["metadata"]["unit"]))

                    if measurement["numeric_value"] is not None:
                        pat_numeric_samples.append(measurement["numeric_value"])

        if is_hierarchical:
            assert numeric_samples is not None
            assert ontology is not None
            final_codes: Set[str] = set()
            for code in code_set:
                final_codes |= ontology.get_all_parents(code)

            for code in final_codes:
                code_counts[code] += 1 / num_patients

            for text in text_set:
                text_counts[text] += 1 / num_patients

            for value in pat_numeric_samples:
                numeric_samples.add(value, 1 / (num_patients * len(pat_numeric_samples)))

    return {
        "age_stats": age_stats,
        "code_counts": code_counts,
        "text_counts": text_counts,
        "numeric_samples": numeric_samples,
        "numeric_samples_by_lab": numeric_samples_by_lab,
    }


def convert_statistics_to_msgpack(
    statistics, vocab_size: int, is_hierarchical: bool, num_numeric: int, ontology: Optional[femr.ontology.Ontology]
):
    vocab = []

    if not is_hierarchical:
        for code, weight in statistics["code_counts"].items():
            entry = {
                "type": "code",
                "code_string": code,
                "weight": weight * math.log(weight) + (1 - weight) * math.log(1 - weight),
            }
            vocab.append(entry)

        for (code, text), weight in statistics["text_counts"].items():
            entry = {
                "type": "text",
                "code_string": code,
                "text_string": text,
                "weight": weight * math.log(weight) + (1 - weight) * math.log(1 - weight),
            }
            vocab.append(entry)

        for code, reservoir in statistics["numeric_samples_by_lab"].items():
            weight = reservoir.total_weight / 10
            samples = reservoir.samples
            samples.sort()

            samples_per_bin = (len(samples) + 9) // 10

            for bin_index in range(0, 10):
                if bin_index == 0:
                    start_val = float("-inf")
                else:
                    if bin_index * samples_per_bin >= len(samples):
                        continue
                    start_val = samples[bin_index * samples_per_bin]

                if bin_index == 9 or (bin_index + 1) * samples_per_bin >= len(samples):
                    end_val = float("inf")
                else:
                    end_val = samples[(bin_index + 1) * samples_per_bin]

                if start_val == end_val:
                    continue

            entry = {
                "type": "numeric",
                "code_string": code,
                "val_start": start_val,
                "val_end": end_val,
                "weight": weight * math.log(weight) + (1 - weight) * math.log(1 - weight),
            }
            vocab.append(entry)
    else:
        assert ontology
        for code, weight in statistics["code_counts"].items():
            baseline = min([1] + [statistics["code_counts"][parent] for parent in ontology.get_parents(code)])
            weight = weight / baseline

            weight = min(1, weight)

            if weight != 0 and weight != 1:
                entry = {
                    "type": "code",
                    "code_string": code,
                    "weight": baseline * (weight * math.log(weight) + (1 - weight) * math.log(1 - weight)),
                }
                vocab.append(entry)

        for text, weight in statistics["text_counts"].items():
            weight = min(1, weight)

            if weight != 0 and weight != 1:
                entry = {
                    "type": "text",
                    "text_string": text,
                    "weight": weight * math.log(weight) + (1 - weight) * math.log(1 - weight),
                }
                vocab.append(entry)

        numeric_samples = list(statistics["numeric_samples"].samples)

        if len(numeric_samples) > 0:
            assert num_numeric >= 1

            samples = sorted(list(set(np.percentile(numeric_samples, np.linspace(0, 1, num=num_numeric + 1)))))
            samples[0] = float("-inf")
            samples[-1] = float("inf")

            assert len(samples) >= 2

            for start_val, end_val in zip(samples, samples[1:]):
                entry = {
                    "type": "numeric",
                    "val_start": start_val,
                    "val_end": end_val,
                    "weight": -1,
                }
                vocab.append(entry)

    vocab.sort(key=lambda a: a["weight"])
    vocab = vocab[:vocab_size]

    result = {
        "vocab": vocab,
        "is_hierarchical": is_hierarchical,
        "age_stats": {
            "mean": statistics["age_stats"].mean(),
            "std": statistics["age_stats"].standard_deviation(),
        },
    }

    return result


class FEMRTokenizer(transformers.utils.PushToHubMixin):
    def __init__(self, dictionary: Mapping[str, Any], ontology: Optional[femr.ontology.Ontology] = None):
        self.dictionary = dictionary

        self.is_hierarchical = dictionary["is_hierarchical"]

        if self.is_hierarchical:
            assert ontology is not None

        self.ontology = ontology

        self.dictionary = dictionary
        vocab = dictionary["vocab"]

        self.string_lookup = {}
        self.code_lookup = {}

        self.vocab_size = len(vocab)

        if not self.is_hierarchical:
            self.numeric_lookup = collections.defaultdict(list)
            for i, dict_entry in enumerate(vocab):
                if dict_entry["type"] == "code":
                    self.code_lookup[dict_entry["code_string"]] = i
                elif dict_entry["type"] == "numeric":
                    self.numeric_lookup[dict_entry["code_string"]].append(
                        (dict_entry["val_start"], dict_entry["val_end"], i)
                    )
                elif dict_entry["type"] == "text":
                    self.string_lookup[(dict_entry["code_string"], dict_entry["text_string"])] = i
                else:
                    pass
        else:
            numeric_entries = []
            for i, dict_entry in enumerate(vocab):
                if dict_entry["type"] == "code":
                    self.code_lookup[dict_entry["code_string"]] = i
                elif dict_entry["type"] == "numeric":
                    numeric_entries.append((dict_entry["val_start"], i))
                elif dict_entry["type"] == "text":
                    self.string_lookup[dict_entry["text_string"]] = i
                else:
                    pass
            numeric_entries.sort()
            if len(numeric_entries) > 0:
                self.numeric_values = [a[0] for a in numeric_entries[1:]]
                self.numeric_indices = [a[1] for a in numeric_entries]
            else:
                self.numeric_values = []
                self.numeric_indices = []

    @classmethod
    def from_pretrained(
        self,
        pretrained_model_name_or_path: Union[str, os.PathLike],
        ontology: Optional[femr.ontology.Ontology] = None,
        **kwargs,
    ):
        """
        Load the FEMR tokenizer.

        Parameters:
            pretrained_model_name_or_path (`str` or `os.PathLike`, *optional*):
                Can be either:
                    - A string, the *model id* of a pretrained model hosted inside a model repo on huggingface.co.
                      Valid model ids can be located at the root-level, like `bert-base-uncased`, or namespaced under a
                      user or organization name, like `dbmdz/bert-base-german-cased`.
                    - A path to a *directory* containing tokenization data saved using
                      [`save_pretrained`], e.g., `./my_data_directory/`.
            ontology: An ontology object for hierarchical tokenizers
            kwargs: Arguments for loading to pass to transformers.utils.hub.cached_file

        Returns:
            A FEMR Tokenizer
        """

        dictionary_file = transformers.utils.hub.cached_file(
            pretrained_model_name_or_path, "dictionary.msgpack", **kwargs
        )

        with open(dictionary_file, "rb") as f:
            dictionary = msgpack.load(f)

        return FEMRTokenizer(dictionary, ontology=ontology)

    def save_pretrained(self, save_directory: Union[str, os.PathLike], push_to_hub: bool = False, **kwargs):
        """
        Save the FEMR tokenizer.


        This method make sure the batch processor can then be re-loaded using the
        .from_pretrained class method.

        Args:
            save_directory (`str` or `os.PathLike`): The path to a directory where the tokenizer will be saved.
            push_to_hub (`bool`, *optional*, defaults to `False`):
                Whether or not to push your model to the Hugging Face model hub after saving it. You can specify the
                repository you want to push to with `repo_id` (will default to the name of `save_directory` in your
                namespace).
            kwargs (`Dict[str, Any]`, *optional*):
                Additional key word arguments passed along to the [`PushToHubMixin.push_to_hub`] method.
        """
        assert not os.path.isfile(save_directory), f"Provided path ({save_directory}) should be a directory, not a file"

        os.makedirs(save_directory, exist_ok=True)

        if push_to_hub:
            commit_message = kwargs.pop("commit_message", None)
            repo_id = kwargs.pop("repo_id", str(save_directory).split(os.path.sep)[-1])
            repo_id = self._create_repo(repo_id, **kwargs)
            files_timestamps = self._get_files_timestamps(save_directory)

        with open(os.path.join(save_directory, "dictionary.msgpack"), "wb") as f:
            msgpack.dump(self.dictionary, f)

        if push_to_hub:
            self._upload_modified_files(
                save_directory,
                repo_id,
                files_timestamps,
                commit_message=commit_message,
                token=kwargs.get("token"),
            )

    def start_patient(self):
        """Compute per-patient statistics that are required to generate features."""

        # This is currently a null-op, but is required for cost featurization
        pass

    def get_feature_codes(self, _time: datetime.datetime, measurement: meds.Measurement) -> Tuple[List[int], Optional[List[float]]]:
        """Get codes for the provided measurement and time"""

        # Note that time is currently not used in this code, but it is required for cost featurization
        if self.is_hierarchical:
            assert self.ontology is not None
            codes = [
                self.code_lookup[parent]
                for parent in self.ontology.get_all_parents(measurement["code"])
                if parent in self.code_lookup
            ]
            weights = [1 / len(codes) for _ in codes]
            if measurement.get("metadata") and normalize_unit(measurement["metadata"].get("unit")) is not None:
                value = self.string_lookup.get(normalize_unit(measurement["metadata"]["unit"]))
                if value is not None:
                    codes.append(value)
                    weights.append(1)
            if measurement.get("numeric_value") is not None and len(self.numeric_indices) > 0:
                codes.append(self.numeric_indices[bisect.bisect(self.numeric_values, measurement["numeric_value"])])
                weights.append(1)
            if measurement.get("text_value") is not None:
                value = self.string_lookup.get(measurement["text_value"])
                if value is not None:
                    codes.append(value)
                    weights.append(1)

            return codes, weights
        else:
            if measurement.get("numeric_value") is not None:
                for start, end, i in self.numeric_lookup.get(measurement["code"], []):
                    if start <= measurement["numeric_value"] < end:
                        return [i], None
                else:
                    return [], None
            elif measurement.get("text_value") is not None:
                value = self.string_lookup.get((measurement["code"], measurement["text_value"]))
                if value is not None:
                    return [value], None
                else:
                    return [], None
            else:
                value = self.code_lookup.get(measurement["code"])
                if value is not None:
                    return [value], None
                else:
                    return [], None

    def normalize_age(self, age: datetime.timedelta) -> float:
        return (age.total_seconds() - self.dictionary["age_stats"]["mean"]) / (self.dictionary["age_stats"]["std"])



"""
A tool for downloading datasets from BigQuery.

Setup:
```
    pip install --upgrade google-cloud-bigquery
    pip install --upgrade google-cloud-storage
    conda install google-cloud-sdk -c conda-forge
```

Note: After installing above packages, run `gcloud auth application-default login` on your terminal.
You will be prompted with a authorization link that you will need to follow and approve using your
email address. Then you will copy-paste authorization code to the terminal.

How to run:
```
python download_bigquery.py <NAME OF YOUR GCP PROJECT> <GCP BIGQUERY DATASET ID> \
    <PATH TO LOCAL FOLDER WHERE DATASET WHERE DATASET WILL BE DOWNLOADED>
    --excluded_tables <(Optional) NAME OF TABLE 1 TO BE IGNORED> <(Optional) NAME OF TABLE 2 TO BE IGNORED>
```

Example: python download_bigquery.py som-nero-nigam-starr \
    som-rit-phi-starr-prod.starr_omop_cdm5_deid_1pcent_lite_2023_02_08 .
"""

from __future__ import annotations

import argparse
import hashlib
import os
import random
import threading
from functools import partial

import google
from google.cloud import bigquery, storage

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Download a Google BigQuery dataset")
    parser.add_argument(
        "gcp_project_name",
        type=str,
        help=(
            "The name of *YOUR* GCP project (e.g. 'som-nero-nigam-starr')."
            " Note that this need NOT be the GCP project that contains the dataset."
            " It just needs to be a GCP project where you have Bucket creation + BigQuery creation permissions."
        ),
    )
    parser.add_argument(
        "gcp_dataset_id",
        type=str,
        help=(
            "The Dataset ID of the GCP dataset to download"
            " (e.g. 'som-rit-phi-starr-prod.starr_omop_cdm5_deid_2022_12_03')."
            " Note that this is the full ID of the dataset (project name + dataset name)"
        ),
    )
    parser.add_argument(
        "output_dir",
        type=str,
        help=(
            "Path to output directory. Note: The downloaded files will be saved in a subdirectory of this,"
            " i.e. `output_dir/gcp_dataset_id/...`"
        ),
    )
    parser.add_argument(
        "--excluded_tables",
        type=str,
        nargs="*",  # 0 or more values expected => creates a list
        default=[],
        help=(
            "Optional. Name(s) of tables to exclude. List tables separated by spaces,"
            " i.e. `--excluded_tables observation note_nlp`"
        ),
    )
    parser.add_argument(
        "--scratch_bucket_postfix",
        type=str,
        default="_extract_scratch",
        help="The postfix for the GCP bucket used for storing temporary files while downloading.",
    )
    args = parser.parse_args()

    target = f"{args.output_dir}/{args.gcp_dataset_id}"
    os.mkdir(target)

    print('Make sure to run "gcloud auth application-default login" before running this command')

    # Connect to our BigQuery project
    client = bigquery.Client(project=args.gcp_project_name)
    storage_client = storage.Client(project=args.gcp_project_name)

    random_dir = hashlib.md5(random.randbytes(16)).hexdigest()

    scratch_bucket_name = args.gcp_project_name.replace("-", "_") + args.scratch_bucket_postfix

    print(f"Storing temporary files in gs://{scratch_bucket_name}/{random_dir}")

    try:
        bucket = storage_client.get_bucket(scratch_bucket_name)
    except google.api_core.exceptions.NotFound as e:
        print(f"Could not find the requested bucket? gs://{scratch_bucket_name} in project {args.gcp_project_name}")
        raise e

    # Get list of all tables in this GCP dataset
    # NOTE: the `HTTPIterator` can be iterated over like a list, but only once (it's a generator)
    tables: google.api_core.page_iterator.HTTPIterator = client.list_tables(args.gcp_dataset_id)
    print(f"Downloading dataset {args.gcp_dataset_id} using your project {args.gcp_project_name}")

    # Use GZIP compression and export as CVSs
    extract_config = bigquery.job.ExtractJobConfig(
        compression=bigquery.job.Compression.GZIP,
        destination_format=bigquery.job.DestinationFormat.CSV,
        field_delimiter=",",
    )

    sem = threading.Semaphore(value=0)  # needed for keeping track of how many tables have been downloaded

    def download(table_id: str, f):
        """Download the results (a set of .csv.gz's) of the BigQuery extract job to our local filesystem
        Note that a single table will be extracted into possibly dozens of smaller .csv.gz files

        Args:
            table_id (str): Name of table (e.g. "attribute_definition")
        """
        if f.errors is not None:
            print("Could not extract, got errors", f.errors, "for", table_id)
            os.abort()
        sem.release()

    n_tables: int = 0
    for table in tables:
        # Get the full name of the table
        table_name: str = f"{table.project}.{table.dataset_id}.{table.table_id}"
        if table.table_id in args.excluded_tables:
            print(f"Skipping extraction | table = {table.table_id}")
            continue
        print(f"Extracting | table = {table.table_id}")
        # Create Google Cloud Storage bucket to extract this table into
        bucket_target_path: str = f"gs://{scratch_bucket_name}/{random_dir}/{table.table_id}/*.csv.gz"
        extract_job = client.extract_table(table.reference, bucket_target_path, job_config=extract_config)
        # Call the `download()` function asynchronously to download the bucket contents to our local filesystem
        extract_job.add_done_callback(partial(download, table.table_id))
        n_tables += 1

    print(f"\n** Extracting a total of {n_tables} tables**\n")
    for i in range(1, n_tables + 1):
        sem.acquire()
        print(f"====> Finished extracting {i} out of {n_tables} tables")

    print("Starting to download tables")

    os.system(f"gsutil -m rsync -r gs://{scratch_bucket_name}/{random_dir} {target}")

    print("------\n------")
    print("Successfully downloaded all tables!")
    print("------\n------")

    # Delete the temporary Google Cloud Storage bucket
    print("\nDeleting temporary files...")
    os.system(f"gsutil -m rm -r gs://{scratch_bucket_name}/{random_dir}")
    print("------\n------")
    print("Successfully deleted temporary Google Cloud Storage files!")
    print("------\n------")



import datetime

import meds


def get_patient_birthdate(patient: meds.Patient) -> datetime.datetime:
    for e in patient["events"]:
        for m in e["measurements"]:
            if m["code"] == meds.birth_code:
                return e["time"]
    raise ValueError("Couldn't find patient birthdate -- Patient has no events " + str(patient["events"][:5]))



import datetime
from typing import Any, List, Optional, Tuple, Union

import datasets
import meds

from femr.labelers import Labeler

# 2nd elem of tuple -- 'skip' means no label, None means censored
EventsWithLabels = List[Tuple[Tuple[Tuple, int, Any], Union[bool, str]]]


DUMMY_EVENTS = [
    ((1995, 1, 3), meds.birth_code, None),
    ((2010, 1, 1), 1, "test_value"),
    ((2010, 1, 1), 1, "test_value"),
    ((2010, 1, 5), 2, 1),
    ((2010, 6, 5), 3, None),
    ((2010, 8, 5), 2, None),
    ((2011, 7, 5), 2, None),
    ((2012, 10, 5), 3, None),
    ((2015, 6, 5, 0), 2, None),
    ((2015, 6, 5, 10, 10), 2, None),
    ((2015, 6, 15, 11), 3, None),
    ((2016, 1, 1), 2, None),
    ((2016, 3, 1, 10, 10, 10), 4, None),
]

NUM_EVENTS = len(DUMMY_EVENTS)
NUM_PATIENTS = 10


def create_patients_dataset(num_patients: int, events: List[Tuple[Tuple, int, Any]] = DUMMY_EVENTS) -> datasets.Dataset:
    """Creates a list of patients, each with the same events contained in `events`"""

    converted_events: List[meds.Event] = []

    for event in events:
        if isinstance(event[1], int):
            code = str(event[1])
        else:
            code = event[1]
        if event[2] is None:
            val = {}
        elif isinstance(event[2], str):
            val = {"text_value": event[2]}
        else:
            val = {"numeric_value": event[2]}
        converted_events.append({"time": datetime.datetime(*event[0]), "measurements": [{"code": code, **val}]})

    return datasets.Dataset.from_dict(
        {"patient_id": list(range(num_patients)), "events": [converted_events for _ in range(num_patients)]}
    )


def assert_labels_are_accurate(
    labeled_patients: List[meds.Label],
    patient_id: int,
    true_labels: List[Tuple[datetime.datetime, Optional[bool]]],
    help_text: str = "",
):
    """Passes if the labels in `labeled_patients` for `patient_id` exactly match the labels in `true_labels`."""
    generated_labels: List[meds.Label] = [a for a in labeled_patients if a["patient_id"] == patient_id]
    # Check that length of lists of labels are the same

    assert len(generated_labels) == len(
        true_labels
    ), f"len(generated): {len(generated_labels)} != len(expected): {len(true_labels)} | {help_text}"
    # Check that value of labels are the same
    for idx, (label, true_label) in enumerate(zip(generated_labels, true_labels)):
        assert label["boolean_value"] == true_label[1] and label["prediction_time"] == true_label[0], (
            f"patient_id={patient_id}, label_idx={idx}, label={label}  |  "
            f"{label} (Assigned) != {true_label} (Expected)  |  "
            f"{help_text}"
        )


def run_test_for_labeler(
    labeler: Labeler,
    events_with_labels: EventsWithLabels,
    true_outcome_times: Optional[List[datetime.datetime]] = None,
    true_prediction_times: Optional[List[datetime.datetime]] = None,
    help_text: str = "",
) -> None:
    patients: datasets.Database = create_patients_dataset(10, [x[0] for x in events_with_labels])
    true_labels: List[Tuple[datetime.datetime, Optional[bool]]] = [
        (datetime.datetime(*x[0][0]), x[1]) for x in events_with_labels if isinstance(x[1], bool)
    ]
    if true_prediction_times is not None:
        # If manually specified prediction times, adjust labels from occurring at `event.start`
        # e.g. we may make predictions at `event.end` or `event.start + 1 day`
        true_labels = [(tp, tl[1]) for (tl, tp) in zip(true_labels, true_prediction_times)]
    labeled_patients: List[meds.Label] = labeler.apply(patients)

    # Check accuracy of Labels
    for patient in patients:
        assert_labels_are_accurate(
            labeled_patients,
            patient["patient_id"],
            true_labels,
            help_text=help_text,
        )



import functools
import pickle


def _agg_helper(*args, map_func):
    result = map_func(*args)
    return {"data": [pickle.dumps(result)]}


def aggregate_over_dataset(dataset, map_func, agg_func, batch_size, num_proc, with_indices=False):
    """Perform an aggregation over a huggingface dataset.

    This logic consists of two parts, map_func and agg_func.

    map_func takes a batch of data and converts it to an intermediate result.

    agg_func takes those intermediate results and combines them into a final result.
    """
    parts = dataset.map(
        functools.partial(_agg_helper, map_func=map_func),
        batched=True,
        batch_size=batch_size,
        remove_columns=dataset.column_names,
        num_proc=num_proc,
        with_indices=with_indices,
        keep_in_memory=True,
        new_fingerprint="invalid",
    )

    current = None
    for stat in parts:
        fixed_stat = pickle.loads(stat["data"])
        if current is None:
            current = fixed_stat
        else:
            current = agg_func(current, fixed_stat)

    return current



"""An ETL script for doing an end to end transform of Stanford data into a PatientDatabase."""

import argparse
import functools
import json
import os
from typing import Callable, Sequence

import datasets
import meds

from femr.transforms import delta_encode, remove_nones
from femr.transforms.stanford import (
    move_billing_codes,
    move_pre_birth,
    move_to_day_end,
    move_visit_start_to_first_event_start,
    switch_to_icd10cm,
)


def _is_visit_measurement(e: meds.Measurement) -> bool:
    return e["metadata"]["table"] == "visit_occurrence"


def _get_stanford_transformations() -> Callable[[meds.Patient], meds.Patient]:
    """Get the list of current OMOP transformations."""
    # All of these transformations are information preserving
    transforms: Sequence[Callable[[meds.Patient], meds.Patient]] = [
        move_pre_birth,
        move_visit_start_to_first_event_start,
        move_to_day_end,
        switch_to_icd10cm,
        move_billing_codes,
        functools.partial(
            remove_nones,  # We have to keep visits in order to sync up visit_ids later in the process
            # If we ever remove or revisit visit_id, we would want to revisit this
            do_not_apply_to_filter=_is_visit_measurement,
        ),
        functools.partial(
            delta_encode,  # We have to keep visits in order to sync up visit_ids later in the process
            # If we ever remove or revisit visit_id, we would want to revisit this
            do_not_apply_to_filter=_is_visit_measurement,
        ),
    ]

    return lambda patient: functools.reduce(lambda r, f: f(r), transforms, patient)


def femr_stanford_omop_fixer_program() -> None:
    """Extract data from an Stanford STARR-OMOP v5 source to create a femr PatientDatabase."""
    parser = argparse.ArgumentParser(description="An extraction tool for STARR-OMOP v5 sources")

    parser.add_argument(
        "source_dataset",
        type=str,
        help="Path of the folder to source dataset",
    )

    parser.add_argument(
        "target_dataset",
        type=str,
        help="The place to store the extract",
    )

    parser.add_argument(
        "--num_proc",
        type=int,
        help="The number of threads to use",
        default=1,
    )

    args = parser.parse_args()

    os.mkdir(args.target_dataset)

    dataset = datasets.Dataset.from_parquet(os.path.join(args.source_dataset, "data", "*"))

    fixed_patient = dataset.map(_get_stanford_transformations(), num_proc=args.num_proc)

    os.mkdir(os.path.join(args.target_dataset, "data"))
    fixed_patient.to_parquet(os.path.join(args.target_dataset, "data", "data.parquet"))

    with open(os.path.join(args.source_dataset, "metadata.json")) as f:
        metadata = json.load(f)

    # Let's mark that we modified this dataset
    metadata["post_etl_name"] = "femr_stanford_omop_fixer"
    metadata["post_etl_version"] = "0.1"

    with open(os.path.join(args.target_dataset, "metadata.json"), "w") as f:
        json.dump(metadata, f)



import torch
import xformers.ops

# From https://github.com/facebookresearch/xformers/blob/042abc8aa47d1f5bcc2e82df041811de218924ba/tests/test_mem_eff_attention.py#L511 # noqa


def ref_attention(q, k, v, attn_bias=None, drop_mask=None, p=0.0, scale=None):
    q = q.float()
    k = k.float()
    v = v.float()

    scale = scale if scale is not None else (1 / q.shape[-1] ** 0.5)
    q = q * scale

    attn = q @ k.transpose(-2, -1)
    if attn_bias is not None:
        if isinstance(attn_bias, xformers.ops.AttentionBias):
            # Always create in B,H,Mq,Mk format
            attn_bias_tensor = attn_bias.materialize(
                (q.shape[0], 1, q.shape[1], k.shape[1]),
                device=q.device,
                dtype=torch.float32,
            )
        else:
            attn_bias_tensor = attn_bias
        if attn_bias_tensor.ndim == 4:
            assert q.shape[0] == attn_bias_tensor.shape[0] * attn_bias_tensor.shape[1]
            attn_bias_tensor = attn_bias_tensor.reshape([-1, *attn_bias_tensor.shape[2:]])
        attn = attn + attn_bias_tensor.float()
    attn = attn.softmax(-1)
    if drop_mask is not None:
        attn = attn * (drop_mask / (1 - p))
    return attn @ v


def ref_attention_bmhk(q, k, v, attn_bias, scale=None) -> torch.Tensor:
    assert q.ndim == 4

    def T(t):
        return t.permute((0, 2, 1, 3)).reshape([t.shape[0] * t.shape[2], t.shape[1], t.shape[3]])

    if isinstance(attn_bias, xformers.ops.AttentionBias):
        attn_bias = attn_bias.materialize(
            (q.shape[0], q.shape[2], q.shape[1], k.shape[1]),
            device=q.device,
            dtype=torch.float32,
        ).reshape([q.shape[0] * q.shape[2], q.shape[1], k.shape[1]])
    out = ref_attention(T(q), T(k), T(v), attn_bias, scale=scale)
    out = out.reshape([q.shape[0], q.shape[2], q.shape[1], v.shape[3]])
    return out.permute((0, 2, 1, 3))


def memory_efficient_attention_wrapper(q, k, v, attn_bias):
    if q.device.type == "cpu":
        return ref_attention_bmhk(q, k, v, attn_bias)
    else:
        return xformers.ops.memory_efficient_attention(q, k, v, attn_bias)



import argparse
import datetime
import json
import os
import pickle
import random

import datasets
import jsonschema
import meds
import pyarrow
import pyarrow.parquet

import femr.ontology
import femr.transforms

parser = argparse.ArgumentParser(prog="generate_patients", description="Create synthetic data")
parser.add_argument("athena", type=str)
parser.add_argument("destination", type=str)
args = parser.parse_args()

random.seed(4533)


def get_random_patient(patient_id):
    epoch = datetime.datetime(1990, 1, 1)
    birth = epoch + datetime.timedelta(days=random.randint(100, 1000))
    current_date = birth

    gender = "Gender/" + random.choice(["F", "M"])
    race = "Race/" + random.choice(["White", "Non-White"])

    patient = {
        "patient_id": patient_id,
        "events": [
            {
                "time": birth,
                "measurements": [
                    {"code": meds.birth_code},
                    {"code": gender},
                    {"code": race},
                ],
            },
        ],
    }
    code_cats = ["ICD9CM", "RxNorm"]
    for code in range(random.randint(1, 10 + (20 if gender == "Gender/F" else 0))):
        code_cat = random.choice(code_cats)
        if code_cat == "RxNorm":
            code = str(random.randint(0, 10000))
        else:
            code = str(random.randint(0, 10000))
            if len(code) > 3:
                code = code[:3] + "." + code[3:]
        current_date = current_date + datetime.timedelta(days=random.randint(1, 100))
        code = code_cat + "/" + code
        patient["events"].append({"time": current_date, "measurements": [{"code": code}]})

    return patient


patients = []
for i in range(200):
    patients.append(get_random_patient(i))

patient_schema = meds.patient_schema()

patient_table = pyarrow.Table.from_pylist(patients, patient_schema)

os.makedirs(os.path.join(args.destination, "data"), exist_ok=True)

pyarrow.parquet.write_table(patient_table, os.path.join(args.destination, "data", "patients.parquet"))

metadata = {
    "dataset_name": "femr synthetic datata",
    "dataset_version": "1",
    "etl_name": "synthetic data",
    "etl_version": "1",
    "code_metadata": {},
}

jsonschema.validate(instance=metadata, schema=meds.dataset_metadata)

with open(os.path.join(args.destination, "metadata.json"), "w") as f:
    json.dump(metadata, f)

dataset = datasets.Dataset.from_parquet(os.path.join(args.destination, "data", "*"))

ontology = femr.ontology.Ontology(args.athena)

ontology.prune_to_dataset(dataset, remove_ontologies=("SNOMED"))

with open(os.path.join(args.destination, "ontology.pkl"), "wb") as f:
    pickle.dump(ontology, f)



from __future__ import annotations

import abc
import collections
import datetime
import functools
from typing import Any, Dict, List, Mapping, Optional, Sequence, Set, Tuple

import datasets
import meds
import numpy as np
import scipy.sparse
import torch

import femr.index
import femr.models.config
import femr.pat_utils
import femr.stat_utils


class Task(abc.ABC):
    def __init__(self):
        super().__init__()

    @abc.abstractmethod
    def get_task_config(self) -> femr.models.config.FEMRTaskConfig:
        ...

    @abc.abstractmethod
    def start_batch(self) -> None:
        ...

    @abc.abstractmethod
    def start_patient(self, patient: meds.Patient, ontology: Optional[femr.ontology.Ontology]) -> None:
        ...

    @abc.abstractmethod
    def needs_exact(self) -> bool:
        ...

    @abc.abstractmethod
    def add_event(
        self, current_date: datetime.datetime, next_date: datetime.datetime, next_features: Sequence[int]
    ) -> int:
        ...

    @abc.abstractmethod
    def get_batch_data(self) -> Mapping[str, np.ndarray]:
        ...

    def cleanup(self, batch: Mapping[str, torch.Tensor]) -> Mapping[str, torch.Tensor]:
        return batch


class LabeledPatientTask(Task):
    def __init__(self, labels: Sequence[meds.Label]):
        super().__init__()

        self.label_map: Mapping[int, Any] = collections.defaultdict(list)
        for label in labels:
            row_without_patient_id = dict(label)
            del row_without_patient_id["patient_id"]
            self.label_map[label["patient_id"]].append(row_without_patient_id)

        for k, v in self.label_map.items():
            v.sort(key=lambda a: a["prediction_time"])

    def get_task_config(self) -> femr.models.config.FEMRTaskConfig:
        return femr.models.config.FEMRTaskConfig(task_type="labeled_patients")

    def filter_dataset(self, dataset: datasets.Dataset, index: femr.index.PatientIndex) -> datasets.Dataset:
        indices = [index.get_index(patient_id) for patient_id in self.label_map]
        return dataset.select(indices)

    def start_patient(self, patient: meds.Patient, _ontology: Optional[femr.ontology.Ontology]) -> None:
        self.current_labels = self.label_map[patient["patient_id"]]
        self.current_label_index = 0

    def needs_exact(self) -> bool:
        return True

    def start_batch(self) -> None:
        """LabeledPatientTask currently has no per label state."""
        pass

    def add_patient_labels(self, _patient_label_offsets: List[int]) -> None:
        """As there is no per label state, this is ignored"""
        pass

    def add_event(
        self, current_date: datetime.datetime, next_date: datetime.datetime, next_features: Optional[Sequence[int]] = None
    ) -> int:
        has_label = False

        while True:
            if self.current_label_index == len(self.current_labels):
                break

            current_label = self.current_labels[self.current_label_index]

            is_valid = current_date <= current_label["prediction_time"]
            next_valid = next_date is not None and next_date <= current_label["prediction_time"]

            if next_valid:
                # Next one is valid, so break eary to give it a chance next time
                break

            if is_valid:
                has_label = True
                self.current_label_index += 1
            else:
                # The next label isn't valid, so we have to break here
                break
        
        if has_label:
            return 1
        else:
            return 0

    def get_batch_data(self) -> Mapping[str, np.ndarray]:
        return {}


class CLMBRTask(Task):
    def __init__(self, clmbr_vocab_size: int):
        self.clmbr_vocab_size = clmbr_vocab_size

    def get_task_config(self) -> femr.models.config.FEMRTaskConfig:
        return femr.models.config.FEMRTaskConfig(
            task_type="clmbr", task_kwargs=dict(clmbr_vocab_size=self.clmbr_vocab_size)
        )

    def start_patient(self, _patient: meds.Patient, _ontology: Optional[femr.ontology.Ontology]) -> None:
        self.per_patient_batch_labels: List[int] = []

    def needs_exact(self) -> bool:
        return False

    def start_batch(self) -> None:
        self.batch_labels: List[int] = []

    def add_patient_labels(self, patient_label_offsets: List[int]) -> None:
        self.batch_labels.extend([self.per_patient_batch_labels[i] for i in patient_label_offsets])

    def add_event(
        self, current_date: datetime.datetime, next_date: datetime.datetime, next_features: Optional[Sequence[int]] = None
    ) -> int:
        if next_features is None:
            return 0

        if len(next_features) != 1:
            raise RuntimeError("Only supports one for right now")

        next_feature = next_features[0]

        if next_feature >= self.clmbr_vocab_size:
            return 0

        self.per_patient_batch_labels.append(next_feature)

        return 1

    def get_batch_data(self) -> Mapping[str, np.ndarray]:
        return {"labels": np.array(self.batch_labels, dtype=np.int32)}


def should_make_survival_prediction(current_date: datetime.datetime, next_date: Optional[datetime.datetime]):
    if next_date is None:
        return False

    if current_date == next_date or current_date.date() == next_date.date():
        return False

    return True


class SurvivalCalculator:
    def __init__(
        self, ontology: femr.ontology.Ontology, patient: meds.Patient, code_whitelist: Optional[Set[str]] = None
    ):
        self.survival_events = []
        self.final_date = patient["events"][-1]["time"]
        self.future_times = collections.defaultdict(list)

        for event in patient["events"]:
            codes = set()
            for measurement in event["measurements"]:
                for parent in ontology.get_all_parents(measurement["code"]):
                    if code_whitelist is None or parent in code_whitelist:
                        codes.add(parent)

            for code in codes:
                self.future_times[code].append(event["time"])
                self.survival_events.append((code, event["time"]))

        for v in self.future_times.values():
            v.reverse()

        self.survival_events.reverse()

    def get_future_events_for_time(
        self, time: datetime.datetime
    ) -> Tuple[datetime.timedelta, Mapping[str, datetime.timedelta]]:
        while len(self.survival_events) > 0 and self.survival_events[-1][1] <= time:
            code = self.survival_events[-1][0]
            vals = self.future_times[code]
            vals.pop()
            if len(vals) == 0:
                del self.future_times[code]

            self.survival_events.pop()

        delta = self.final_date - time
        return (delta, {k: v[-1] - time for k, v in self.future_times.items()})


def _prefit_motor_map(batch, *, tasks: List[str], ontology: femr.ontology.Ontology) -> Any:
    task_time_stats: List[Any] = [[0, 0, femr.stat_utils.OnlineStatistics()] for _ in range(len(tasks))]
    event_times = femr.stat_utils.ReservoirSampler(100_000)
    task_set = set(tasks)

    for patient_id, events in zip(batch["patient_id"], batch["events"]):
        patient = {"patient_id": patient_id, "events": events}

        calculator = SurvivalCalculator(ontology, patient, task_set)

        birth = femr.pat_utils.get_patient_birthdate(patient)

        for event, next_event in zip(patient["events"], patient["events"][1:]):
            if (event["time"] - birth).days <= 1:
                continue
            if should_make_survival_prediction(event["time"], next_event["time"]):
                censor_time, tte = calculator.get_future_events_for_time(event["time"])

                for i, task in enumerate(tasks):
                    if task in tte:
                        time = tte[task]
                        is_censored = False
                    else:
                        time = censor_time
                        is_censored = True

                    if is_censored:
                        task_time_stats[i][0] += 1
                    else:
                        event_times.add(time.total_seconds(), 1)
                        task_time_stats[i][1] += 1
                    task_time_stats[i][2].add(1, time.total_seconds())

    return (event_times, task_time_stats)


def _prefit_motor_agg(first: Any, second: Any) -> Any:
    for a, b in zip(first[1], second[1]):
        a[0] += b[0]
        a[1] += b[1]
        a[2].combine(b[2])
    first[0].combine(second[0])
    return first


class MOTORTask(Task):
    @classmethod
    def fit_pretraining_task_info(
        cls,
        dataset: datasets.Dataset,
        tokenizer: femr.models.tokenizer.FEMRTokenizer,
        num_tasks: int,
        num_bins: int,
        final_layer_size: int,
        num_proc: int = 1,
    ) -> MOTORTask:
        tasks = []
        for dict_entry in tokenizer.dictionary["vocab"]:
            if dict_entry["type"] == "code":
                tasks.append(dict_entry["code_string"])
                if len(tasks) == num_tasks:
                    break

        assert len(tasks) == num_tasks, "Could not find enough tasks in the provided tokenizer"

        length_samples, stats = femr.hf_utils.aggregate_over_dataset(
            dataset,
            functools.partial(_prefit_motor_map, tasks=tasks, ontology=tokenizer.ontology),
            _prefit_motor_agg,
            1_000,
            num_proc=num_proc,
        )

        time_bins = np.percentile(length_samples.samples, np.linspace(0, 100, num_bins + 1))
        time_bins[0] = 0
        time_bins[-1] = float("inf")
        time_bins = list(time_bins)

        task_data = []

        for task, task_stats in zip(tasks, stats):
            frac_events = task_stats[1] / (task_stats[0] + task_stats[1])
            rate = frac_events / task_stats[2].mean()
            task_data.append((task, rate))

        return MOTORTask(task_data, time_bins, final_layer_size)

    def __init__(self, pretraining_task_info: List[Tuple[str, float]], time_bins: List[float], final_layer_size: int):
        self.pretraining_task_info = pretraining_task_info
        self.time_bins = time_bins
        self.final_layer_size = final_layer_size

        self.pretraining_task_codes = set()
        self.task_to_index_map = {}
        for i, task in enumerate(self.pretraining_task_info):
            self.pretraining_task_codes.add(task[0])
            self.task_to_index_map[task[0]] = i

    def get_task_config(self) -> femr.models.config.FEMRTaskConfig:
        return femr.models.config.FEMRTaskConfig(
            task_type="motor",
            task_kwargs=dict(
                pretraining_task_info=self.pretraining_task_info,
                time_bins=self.time_bins,
                final_layer_size=self.final_layer_size,
            ),
        )

    def start_patient(self, patient: meds.Patient, ontology: Optional[femr.ontology.Ontology]) -> None:
        assert ontology
        self.calculator = SurvivalCalculator(ontology, patient, self.pretraining_task_codes)

        self.per_patient_censor_time = []
        self.per_patient_time_sparse = {
            "data": [],
            "indices": [],
            "indptr": [0],
        }

    def needs_exact(self) -> bool:
        return False

    def start_batch(self) -> None:
        self.censor_time: List[float] = []

        self.time_sparse: Dict[str, List[float]] = {
            "data": [],
            "indices": [],
            "indptr": [0],
        }

    def add_patient_labels(self, patient_label_offsets: List[int]) -> None:
        """Add per-patient labels to the global task labels."""
        self.censor_time.extend([self.per_patient_censor_time[i] for i in patient_label_offsets])

        for index in patient_label_offsets:
            start = self.per_patient_time_sparse['indptr'][index]
            end = self.per_patient_time_sparse['indptr'][index + 1]

            self.time_sparse['data'].extend(self.per_patient_time_sparse['data'][start:end])
            self.time_sparse['indices'].extend(self.per_patient_time_sparse['indices'][start:end])
            self.time_sparse['indptr'].append(len(self.time_sparse['indices']))

    def add_event(
        self, current_date: datetime.datetime, next_date: datetime.datetime, next_features: Sequence[int]
    ) -> int:
        if not should_make_survival_prediction(current_date, next_date):
            return 0

        censor_time, tte = self.calculator.get_future_events_for_time(current_date)

        if len(tte) == 0:
            return 0

        censor_seconds = censor_time.total_seconds()
        self.per_patient_censor_time.append(censor_seconds)

        for event_name, time in tte.items():
            j = self.task_to_index_map[event_name]
            seconds = time.total_seconds()

            self.per_patient_time_sparse["data"].append(seconds)
            self.per_patient_time_sparse["indices"].append(j)

        self.per_patient_time_sparse["indptr"].append(len(self.per_patient_time_sparse["data"]))

        return 1

    def get_batch_data(self) -> Mapping[str, np.ndarray]:
        def h(a, dtype):
            return {
                "data": np.array(a["data"], dtype=dtype),
                "indices": np.array(a["indices"], dtype=np.int32),
                "indptr": np.array(a["indptr"], dtype=np.int32),
            }

        return {
            "censor_time": np.array(self.censor_time, dtype=np.float32),
            "time_sparse": h(self.time_sparse, dtype=np.float32),
        }

    def cleanup(self, batch: Mapping[str, torch.Tensor]) -> Mapping[str, torch.Tensor]:
        num_time_bins = len(self.time_bins) - 1
        num_tasks = len(self.pretraining_task_info)
        num_indices = len(batch["censor_time"])

        def h(a):
            shape = (num_indices, num_tasks)
            a = {k: v.numpy() for k, v in batch[a].items()}
            s = scipy.sparse.csr_array((a["data"], a["indices"], a["indptr"]), shape=shape)
            return torch.from_numpy(s.toarray())

        time = h("time_sparse")

        log_time = torch.zeros(size=(num_time_bins, num_indices, num_tasks), dtype=torch.float16)
        is_event = torch.zeros(size=(num_time_bins, num_indices, num_tasks), dtype=torch.bool)

        is_event_global = time != 0

        def inf_log(arr):
            mask = arr != 0
            arr[mask] = torch.log2(arr[mask])
            arr[~mask] = -torch.inf

            return arr

        for i, (start, end) in enumerate(zip(self.time_bins, self.time_bins[1:])):
            censor_time_in_bin = torch.unsqueeze(torch.clip(batch["censor_time"] - start, 0, float(end - start)), -1)
            event_time_in_bin = torch.clip(time - start, 0, float(end - start))
            time_in_bin = torch.where(is_event_global, event_time_in_bin, censor_time_in_bin)
            log_time[i, :] = inf_log(time_in_bin)
            is_event[i, :] = is_event_global & (start <= time) & (time < end)

        log_time = torch.transpose(log_time, 0, 1).contiguous()
        is_event = torch.transpose(is_event, 0, 1).contiguous()

        return {"is_event": is_event, "log_time": log_time}



"""Core labeling functionality/schemas, shared across all labeling functions."""
from __future__ import annotations

import datetime
import functools
import hashlib
import struct
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional, Tuple

import datasets
import meds

import femr.hf_utils


@dataclass(frozen=True)
class TimeHorizon:
    """An interval of time. Mandatory `start`, optional `end`."""

    start: datetime.timedelta
    end: datetime.timedelta | None  # If NONE, then infinite time horizon


def _label_map_func(batch, *, labeler: Labeler) -> List[meds.Label]:
    result = []
    for patient_id, events in zip(batch["patient_id"], batch["events"]):
        result.extend(labeler.label({"patient_id": patient_id, "events": events}))
    return result


def _label_agg_func(first_labels: List[meds.Label], second_labels: List[meds.Label]):
    first_labels.extend(second_labels)

    return first_labels


class Labeler(ABC):
    """An interface for labeling functions.

    A labeling function applies a label to a specific datetime in a given patient's timeline.
    It can be thought of as generating the following list given a specific patient:
        [(patient ID, datetime_1, label_1), (patient ID, datetime_2, label_2), ... ]
    Usage:
    ```
        labeling_function: Labeler = Labeler(...)
        patients: Sequence[Patient] = ...
        labels: LabeledPatient = labeling_function.apply(patients)
    ```
    """

    @abstractmethod
    def label(self, patient: meds.Patient) -> List[meds.Label]:
        """Apply every label that is applicable to the provided patient.

        This is only called once per patient.

        Args:
            patient (Patient): A patient object

        Returns:
            List[Label]: A list of :class:`Label` containing every label for the given patient
        """
        pass

    def apply(
        self,
        dataset: datasets.Dataset,
        num_proc: int = 1,
    ) -> List[meds.Label]:
        """Apply the `label()` function one-by-one to each Patient in a sequence of Patients.

        Args:
            dataset (datasets.Dataset): A HuggingFace Dataset with meds.Patient objects to be labeled.
            num_proc (int, optional): Number of CPU threads to parallelize across. Defaults to 1.

        Returns:
            A list of labels
        """

        return femr.hf_utils.aggregate_over_dataset(
            dataset,
            functools.partial(_label_map_func, labeler=self),
            _label_agg_func,
            batch_size=10_000,
            num_proc=num_proc,
        )


##########################################################
# Specific Labeler Superclasses
##########################################################


class TimeHorizonEventLabeler(Labeler):
    """Label events that occur within a particular time horizon.
    This support both "finite" and "infinite" time horizons.

    The time horizon can be "fixed" (i.e. has both a start and end date), or "infinite" (i.e. only a start date)

    A TimeHorizonEventLabeler enables you to label events that occur within a particular
    time horizon (i.e. `TimeHorizon`). It is a boolean event that is TRUE if the event of interest
    occurs within that time horizon, and FALSE if it doesn't occur by the end of the time horizon.

    No labels are generated if the patient record is "censored" before the end of the horizon.

    You are required to implement three methods:
        get_outcome_times() for defining the datetimes of the event of interset
        get_prediction_times() for defining the datetimes at which we make our predictions
        get_time_horizon() for defining the length of time (i.e. `TimeHorizon`) to use for the time horizon
    """

    def __init__(self):
        pass

    @abstractmethod
    def get_outcome_times(self, patient: meds.Patient) -> List[datetime.datetime]:
        """Return a sorted list containing the datetimes that the event of interest "occurs".

        IMPORTANT: Must be sorted ascending (i.e. start -> end of timeline)

        Args:
            patient (Patient): A patient object

        Returns:
            List[datetime.datetime]: A list of datetimes, one corresponding to an occurrence of the outcome
        """
        pass

    @abstractmethod
    def get_time_horizon(self) -> TimeHorizon:
        """Return time horizon for making predictions with this labeling function.

        Return the (start offset, end offset) of the time horizon (from the prediction time)
        used for labeling whether an outcome occurred or not. These can be arbitrary timedeltas.

        If end offset is None, then the time horizon is infinite (i.e. only has a start offset).
        If end offset is not None, then the time horizon is finite (i.e. has both a start and end offset),
            and it must be true that end offset >= start offset.

        Example:
            X is the time that you're making a prediction (given by `get_prediction_times()`)
            (A,B) is your time horizon (given by `get_time_horizon()`)
            O is an outcome (given by `get_outcome_times()`)

            Then given a patient timeline:
                X-----(X+A)------(X+B)------


            This has a label of TRUE:
                X-----(X+A)--O---(X+B)------

            This has a label of TRUE:
                X-----(X+A)--O---(X+B)----O-

            This has a label of FALSE:
                X---O-(X+A)------(X+B)------

            This has a label of FALSE:
                X-----(X+A)------(X+B)--O---
        """
        pass

    @abstractmethod
    def get_prediction_times(self, patient: meds.Patient) -> List[datetime.datetime]:
        """Return a sorted list containing the datetimes at which we'll make a prediction.

        IMPORTANT: Must be sorted ascending (i.e. start -> end of timeline)
        """
        pass

    def get_patient_start_end_times(self, patient: meds.Patient) -> Tuple[datetime.datetime, datetime.datetime]:
        """Return the datetimes that we consider the (start, end) of this patient."""
        return (patient["events"][0]["time"], patient["events"][-1]["time"])

    def allow_same_time_labels(self) -> bool:
        """Whether or not to allow labels with events at the same time as prediction"""
        return True

    def label(self, patient: meds.Patient) -> List[meds.Label]:
        """Return a list of Labels for an individual patient.

        Assumes that events in `patient['events']` are already sorted in chronologically
        ascending order (i.e. start -> end).

        Args:
            patient (Patient): A patient object

        Returns:
            List[Label]: A list containing a label for each datetime returned by `get_prediction_times()`
        """
        if len(patient["events"]) == 0:
            return []

        __, end_time = self.get_patient_start_end_times(patient)
        outcome_times: List[datetime.datetime] = self.get_outcome_times(patient)
        prediction_times: List[datetime.datetime] = self.get_prediction_times(patient)
        time_horizon: TimeHorizon = self.get_time_horizon()

        # Get (start, end) of time horizon. If end is None, then it's infinite (set timedelta to max)
        time_horizon_start: datetime.timedelta = time_horizon.start
        time_horizon_end: Optional[datetime.timedelta] = time_horizon.end  # `None` if infinite time horizon

        # For each prediction time, check if there is an outcome which occurs within the (start, end)
        # of the time horizon
        results: List[meds.Label] = []
        curr_outcome_idx: int = 0
        last_time = None

        for time in prediction_times:
            if last_time is not None:
                assert time > last_time, f"Must be ascending prediction times, instead got {last_time} <= {time}"

            last_time = time
            while curr_outcome_idx < len(outcome_times) and outcome_times[curr_outcome_idx] < time + time_horizon_start:
                # `curr_outcome_idx` is the idx in `outcome_times` that corresponds to the first
                # outcome EQUAL or AFTER the time horizon for this prediction time starts (if one exists)
                curr_outcome_idx += 1

            if curr_outcome_idx < len(outcome_times) and outcome_times[curr_outcome_idx] == time:
                if not self.allow_same_time_labels():
                    continue
                warnings.warn(
                    "You are making predictions at the same time as the target outcome."
                    "This frequently leads to label leakage."
                )

            # TRUE if an event occurs within the time horizon
            is_outcome_occurs_in_time_horizon: bool = (
                (
                    # ensure there is an outcome
                    # (needed in case there are 0 outcomes)
                    curr_outcome_idx
                    < len(outcome_times)
                )
                and (
                    # outcome occurs after time horizon starts
                    time + time_horizon_start
                    <= outcome_times[curr_outcome_idx]
                )
                and (
                    # outcome occurs before time horizon ends (if there is an end)
                    (time_horizon_end is None)
                    or outcome_times[curr_outcome_idx] <= time + time_horizon_end
                )
            )
            # TRUE if patient is censored (i.e. timeline ends BEFORE this time horizon ends,
            # so we don't know if the outcome happened after the patient timeline ends)
            # If infinite time horizon labeler, then assume no censoring
            is_censored: bool = end_time < time + time_horizon_end if (time_horizon_end is not None) else False

            if is_outcome_occurs_in_time_horizon:
                results.append(meds.Label(patient_id=patient["patient_id"], prediction_time=time, boolean_value=True))
            elif not is_censored:
                # Not censored + no outcome => FALSE
                results.append(meds.Label(patient_id=patient["patient_id"], prediction_time=time, boolean_value=False))
            elif is_censored:
                # Censored => None
                pass

        return results


class NLabelsPerPatientLabeler(Labeler):
    """Restricts `self.labeler` to returning a max of `self.k` labels per patient."""

    def __init__(self, labeler: Labeler, num_labels: int = 1, seed: int = 1):
        self.labeler: Labeler = labeler
        self.num_labels: int = num_labels  # number of labels per patient
        self.seed: int = seed

    def label(self, patient: meds.Patient) -> List[meds.Label]:
        labels: List[meds.Label] = self.labeler.label(patient)
        if len(labels) <= self.num_labels:
            return labels
        elif self.num_labels == -1:
            return labels
        hash_to_label_list: List[Tuple[int, int, meds.Label]] = [
            (i, compute_random_num(self.seed, patient["patient_id"], i), labels[i]) for i in range(len(labels))
        ]
        hash_to_label_list.sort(key=lambda a: a[1])
        n_hash_to_label_list: List[Tuple[int, int, meds.Label]] = hash_to_label_list[: self.num_labels]
        n_hash_to_label_list.sort(key=lambda a: a[0])
        n_labels: List[meds.Label] = [hash_to_label[2] for hash_to_label in n_hash_to_label_list]
        return n_labels


def compute_random_num(seed: int, num_1: int, num_2: int, modulus: int = 100):
    network_num_1 = struct.pack("!q", num_1)
    network_num_2 = struct.pack("!q", num_2)
    network_seed = struct.pack("!q", seed)

    to_hash = network_seed + network_num_1 + network_num_2

    hash_object = hashlib.sha256()
    hash_object.update(to_hash)
    hash_value = hash_object.digest()

    result = 0
    for i in range(len(hash_value)):
        result = (result * 256 + hash_value[i]) % modulus

    return result



from __future__ import annotations

import datasets

import femr.hf_utils


def map_index(batch, indices):
    return list(zip(batch["patient_id"], indices))


def agg_index(index1, index2):
    index1.extend(index2)
    return index1


class PatientIndex:
    def __init__(self, dataset: datasets.Dataset, num_proc: int = 1):
        data = femr.hf_utils.aggregate_over_dataset(
            dataset, map_index, agg_index, num_proc=num_proc, batch_size=1_000, with_indices=True
        )
        self.index_map = dict(data)

    def get_patient_ids(self):
        return self.index_map.keys()

    def get_index(self, patient_id):
        return self.index_map[patient_id]

    def filter_dataset(self, dataset, patient_ids):
        return dataset.select([self.get_index(patient_id) for patient_id in patient_ids])



from __future__ import annotations

import datetime

import meds

from femr.transforms import delta_encode, remove_nones
from femr.transforms.stanford import (
    move_billing_codes,
    move_pre_birth,
    move_to_day_end,
    move_visit_start_to_first_event_start,
)


def cleanup(patient):
    for event in patient["events"]:
        for measurement in event["measurements"]:
            if "metadata" not in measurement:
                measurement["metadata"] = {}

            for k in ("numeric_value", "text_value", "datetime_value"):
                if k not in measurement:
                    measurement[k] = None

            if "table" not in measurement["metadata"]:
                measurement["metadata"]["table"] = None


def test_pre_birth() -> None:
    patient = {
        "patient_id": 123,
        "events": [
            {"time": datetime.datetime(1999, 7, 2), "measurements": [{"code": 1234}]},
            {"time": datetime.datetime(1999, 7, 9), "measurements": [{"code": meds.birth_code}]},
            {"time": datetime.datetime(1999, 7, 11), "measurements": [{"code": 12345}]},
        ],
    }

    expected = {
        "patient_id": 123,
        "events": [
            {"time": datetime.datetime(1999, 7, 9), "measurements": [{"code": 1234}]},
            {"time": datetime.datetime(1999, 7, 9), "measurements": [{"code": meds.birth_code}]},
            {"time": datetime.datetime(1999, 7, 11), "measurements": [{"code": 12345}]},
        ],
    }

    cleanup(patient)
    cleanup(expected)

    assert move_pre_birth(patient) == expected


def test_move_visit_start_ignores_other_visits() -> None:
    patient = {
        "patient_id": 123,
        "events": [
            {  # A non-visit event with no explicit start time
                "time": datetime.datetime(1999, 7, 2),
                "measurements": [{"code": 1234, "metadata": {"visit_id": 9999}}],
            },
            {  # A visit event with just date specified
                "time": datetime.datetime(1999, 7, 2),
                "measurements": [
                    {
                        "code": 4567,
                        "metadata": {
                            "table": "visit_occurrence",
                            "visit_id": 9999,
                        },
                    }
                ],
            },
            {  # A non-visit event from a separate visit ID
                "time": datetime.datetime(1999, 7, 2, 11),
                "measurements": [{"code": 2345, "metadata": {"visit_id": 8888}}],
            },
            {  # First recorded non-visit event for visit ID 9999
                "time": datetime.datetime(1999, 7, 2, 12),
                "measurements": [{"code": 3456, "metadata": {"visit_id": 9999}}],
            },
        ],
    }

    # Note that events are implicitly sorted first by start time, then by code:
    # https://github.com/som-shahlab/femr/blob/main/src/femr/__init__.py#L69
    expected = {
        "patient_id": 123,
        "events": [
            {  # A non-visit event with no explicit start time
                "time": datetime.datetime(1999, 7, 2),
                "measurements": [{"code": 1234, "metadata": {"visit_id": 9999}}],
            },
            {  # A non-visit event from a separate visit ID
                "time": datetime.datetime(1999, 7, 2, 11),
                "measurements": [{"code": 2345, "metadata": {"visit_id": 8888}}],
            },
            {  # Now visit event has date and time specified
                "time": datetime.datetime(1999, 7, 2, 12),
                "measurements": [
                    {
                        "code": 4567,
                        "metadata": {
                            "table": "visit_occurrence",
                            "visit_id": 9999,
                        },
                    }
                ],
            },
            {  # First recorded non-visit event for visit ID 9999
                "time": datetime.datetime(1999, 7, 2, 12),
                "measurements": [{"code": 3456, "metadata": {"visit_id": 9999}}],
            },
        ],
    }

    cleanup(patient)
    cleanup(expected)

    assert move_visit_start_to_first_event_start(patient) == expected


def test_move_visit_start_minute_after_midnight() -> None:
    patient = {
        "patient_id": 123,
        "events": [
            {
                "time": datetime.datetime(1999, 7, 2),
                "measurements": [
                    {"code": 3456, "metadata": {"visit_id": 9999, "table": "visit_occurrence"}},
                    {"code": 1234, "metadata": {"visit_id": 9999}},
                ],
            },
            {
                "time": datetime.datetime(1999, 7, 2, 0, 1),
                "measurements": [{"code": 2345, "metadata": {"visit_id": 9999}}],
            },
            {
                "time": datetime.datetime(1999, 7, 2, 12),
                "measurements": [{"code": 4567, "metadata": {"visit_id": 9999}}],
            },
        ],
    }

    expected = {
        "patient_id": 123,
        "events": [
            {
                "time": datetime.datetime(1999, 7, 2),
                "measurements": [{"code": 1234, "metadata": {"visit_id": 9999}}],
            },
            {
                "time": datetime.datetime(1999, 7, 2, 0, 1),
                "measurements": [{"code": 3456, "metadata": {"visit_id": 9999, "table": "visit_occurrence"}}],
            },
            {
                "time": datetime.datetime(1999, 7, 2, 0, 1),
                "measurements": [{"code": 2345, "metadata": {"visit_id": 9999}}],
            },
            {
                "time": datetime.datetime(1999, 7, 2, 12),
                "measurements": [{"code": 4567, "metadata": {"visit_id": 9999}}],
            },
        ],
    }

    cleanup(patient)
    cleanup(expected)

    assert move_visit_start_to_first_event_start(patient) == expected


def test_move_visit_start_doesnt_move_without_event() -> None:
    patient = {
        "patient_id": 123,
        "events": [
            {
                "time": datetime.datetime(1999, 7, 2),
                "measurements": [
                    {"code": 1234, "metadata": {"visit_id": 9999}},
                    {"code": 3456, "metadata": {"visit_id": 9999, "table": "visit_occurrence"}},
                    {"code": 2345, "metadata": {"visit_id": 9999}},
                ],
            },
        ],
    }

    # None of the non-visit events have start time > '00:00:00' so visit event
    # start time is unchanged, though order changes based on code under resort.
    expected = {
        "patient_id": 123,
        "events": [
            {
                "time": datetime.datetime(1999, 7, 2),
                "measurements": [
                    {"code": 1234, "metadata": {"visit_id": 9999}},
                    {"code": 3456, "metadata": {"visit_id": 9999, "table": "visit_occurrence"}},
                    {"code": 2345, "metadata": {"visit_id": 9999}},
                ],
            }
        ],
    }

    cleanup(patient)
    cleanup(expected)

    assert move_visit_start_to_first_event_start(patient) == expected


def test_move_to_day_end() -> None:
    patient = {
        "patient_id": 123,
        "events": [
            {"time": datetime.datetime(1999, 7, 2), "measurements": [{"code": 1234}]},
            {"time": datetime.datetime(1999, 7, 2, 12), "measurements": [{"code": 4321}]},
            {"time": datetime.datetime(1999, 7, 9), "measurements": [{"code": meds.birth_code}]},
        ],
    }

    expected = {
        "patient_id": 123,
        "events": [
            {"time": datetime.datetime(1999, 7, 2, 12), "measurements": [{"code": 4321}]},
            {"time": datetime.datetime(1999, 7, 2, 23, 59), "measurements": [{"code": 1234}]},
            {"time": datetime.datetime(1999, 7, 9, 23, 59), "measurements": [{"code": meds.birth_code}]},
        ],
    }

    cleanup(patient)
    cleanup(expected)

    print(move_to_day_end(patient))
    print(expected)

    assert move_to_day_end(patient) == expected


def test_remove_nones() -> None:
    patient = {
        "patient_id": 123,
        "events": [
            {"time": datetime.datetime(1999, 7, 2), "measurements": [{"code": 1234}]},  # No value, to be removed
            {"time": datetime.datetime(1999, 7, 2, 12), "measurements": [{"code": 1234, "numeric_value": 3}]},
            {"time": datetime.datetime(1999, 7, 9), "measurements": [{"code": meds.birth_code}]},
        ],
    }

    expected = {
        "patient_id": 123,
        "events": [
            {"time": datetime.datetime(1999, 7, 2, 12), "measurements": [{"code": 1234, "numeric_value": 3}]},
            {"time": datetime.datetime(1999, 7, 9), "measurements": [{"code": meds.birth_code}]},
        ],
    }

    cleanup(patient)
    cleanup(expected)

    assert remove_nones(patient) == expected


def test_delta_encode() -> None:
    patient = {
        "patient_id": 123,
        "events": [
            {"time": datetime.datetime(1999, 7, 2), "measurements": [{"code": 1234}]},
            {"time": datetime.datetime(1999, 7, 2), "measurements": [{"code": 1234}]},
            {"time": datetime.datetime(1999, 7, 2, 12), "measurements": [{"code": 1234, "numeric_value": 3}]},
            {"time": datetime.datetime(1999, 7, 2, 14), "measurements": [{"code": 1234, "numeric_value": 3}]},
            {"time": datetime.datetime(1999, 7, 2, 19), "measurements": [{"code": 1234, "numeric_value": 5}]},
            {"time": datetime.datetime(1999, 7, 2, 20), "measurements": [{"code": 1234, "numeric_value": 3}]},
        ],
    }

    expected = {
        "patient_id": 123,
        "events": [
            {"time": datetime.datetime(1999, 7, 2), "measurements": [{"code": 1234}]},
            {"time": datetime.datetime(1999, 7, 2, 12), "measurements": [{"code": 1234, "numeric_value": 3}]},
            {"time": datetime.datetime(1999, 7, 2, 19), "measurements": [{"code": 1234, "numeric_value": 5}]},
            {"time": datetime.datetime(1999, 7, 2, 20), "measurements": [{"code": 1234, "numeric_value": 3}]},
        ],
    }

    cleanup(patient)
    cleanup(expected)

    assert delta_encode(patient) == expected


def test_move_billing_codes() -> None:
    patient = {
        "patient_id": 123,
        "events": [
            {
                "time": datetime.datetime(1999, 7, 2),
                "measurements": [
                    {
                        "code": 1234,
                        "metadata": {
                            "end": datetime.datetime(1999, 7, 20),
                            "visit_id": 10,
                            "clarity_table": "lpch_pat_enc",
                        },
                    }
                ],
            },
            {
                "time": datetime.datetime(1999, 7, 9),
                "measurements": [
                    {
                        "code": meds.birth_code,
                        "metadata": {
                            "visit_id": 10,
                            "clarity_table": "lpch_pat_enc_dx",
                        },
                    }
                ],
            },
            {
                "time": datetime.datetime(1999, 7, 10),
                "measurements": [
                    {
                        "code": 42165,
                        "metadata": {
                            "visit_id": 10,
                            "clarity_table": "shc_pat_enc_dx",
                        },
                    }
                ],
            },
            {
                "time": datetime.datetime(1999, 7, 11),
                "measurements": [
                    {
                        "code": 12345,
                        "metadata": {
                            "visit_id": 10,
                        },
                    }
                ],
            },
            {
                "time": datetime.datetime(1999, 7, 13),
                "measurements": [
                    {
                        "code": 123,
                        "metadata": {
                            "visit_id": 11,
                        },
                    }
                ],
            },
        ],
    }

    expected = {
        "patient_id": 123,
        "events": [
            {
                "time": datetime.datetime(1999, 7, 2),
                "measurements": [
                    {
                        "code": 1234,
                        "metadata": {
                            "end": datetime.datetime(1999, 7, 20),
                            "visit_id": 10,
                            "clarity_table": "lpch_pat_enc",
                        },
                    }
                ],
            },
            {
                "time": datetime.datetime(1999, 7, 11),
                "measurements": [
                    {
                        "code": 12345,
                        "metadata": {
                            "visit_id": 10,
                        },
                    }
                ],
            },
            {
                "time": datetime.datetime(1999, 7, 13),
                "measurements": [
                    {
                        "code": 123,
                        "metadata": {
                            "visit_id": 11,
                        },
                    }
                ],
            },
            {
                "time": datetime.datetime(1999, 7, 20),
                "measurements": [
                    {
                        "code": meds.birth_code,
                        "metadata": {
                            "visit_id": 10,
                            "clarity_table": "lpch_pat_enc_dx",
                        },
                    }
                ],
            },
            {
                "time": datetime.datetime(1999, 7, 20),
                "measurements": [
                    {
                        "code": 42165,
                        "metadata": {
                            "visit_id": 10,
                            "clarity_table": "shc_pat_enc_dx",
                        },
                    }
                ],
            },
        ],
    }

    cleanup(patient)
    cleanup(expected)

    print(move_billing_codes(patient))
    print(expected)

    assert move_billing_codes(patient) == expected
